"""Superlattice analysis tools for charge-density wave (CDW) and spin-density wave (SDW) computations.

The :class:`superlattice` class folds lattice sites into a supercell, computes CDW/SDW
eigenstates and energies, and provides visualization of CDW patterns and interaction links.
Intended for use with the inter-cluster Hartree mean-field approximation.
"""
import numpy as np
import matplotlib.pyplot as plt

np.set_printoptions(precision=10, linewidth=512, suppress=True)

####################################################################################################
class superlattice:
    """
    Class used to analyze superlattices in view of the inter-cluster Hartree mean-field approximation
    """
    def __init__(self, sites, e):
        """
        This class is used to define a supercluster and the associated superlattice
        :param list[list[int]] sites: list of sites of the cluster (integer components), or tuple of such collections if more than one cluster in a supercluster.
        :param list[list[int]] e: superlattice basis vectors (integer components, the number of vectors is the dimension). Must be right oriented.
        """
        e = np.array(e, dtype='int')
        self.e = e
        self.dim = self.e.shape[0]
        assert (self.dim > 0 and self.dim <= 3), 'the dimension must be 1, 2 or 3'
        self.E = np.eye(3, dtype='int')
        self.E[0:self.dim, :] = self.e
        S = self.E
        self.vol = int(np.rint(np.linalg.det(S)))  # fixed : go to nearest integer
        assert self.vol > 0, 'Please use a lattice with a positive orientation. Aborting'

        self.minors = np.empty((3,3), dtype=int)
        self.minors[0,0] =  S[1,1]*S[2,2] - S[2,1]*S[1,2]
        self.minors[1,0] = -S[1,0]*S[2,2] + S[2,0]*S[1,2]
        self.minors[2,0] =  S[1,0]*S[2,1] - S[1,1]*S[2,0]
        self.minors[0,1] = -S[0,1]*S[2,2] + S[2,1]*S[0,2]
        self.minors[1,1] =  S[0,0]*S[2,2] - S[2,0]*S[0,2]
        self.minors[2,1] = -S[0,0]*S[2,1] + S[0,1]*S[2,0]
        self.minors[0,2] =  S[0,1]*S[1,2] - S[1,1]*S[0,2]
        self.minors[1,2] = -S[0,0]*S[1,2] + S[1,0]*S[0,2]
        self.minors[2,2] =  S[0,0]*S[1,1] - S[0,1]*S[1,0]
        # print('minors:\n', self.minors)
        # print('super unit cell volume = ', self.vol)

        # folding the sites
        self.siteF_index = {}
        if type(sites) is not tuple:
            sites = (sites,)
        sitesF = []
        sitesO = []
        clus=[]
        for i,c in enumerate(sites):
            csites = np.array(c, dtype='int')
            for x in csites:
                sitesF.append(self.fold_shifted(x)[1])
                sitesO.append(x)
                clus.append(i)
        self.sitesF = np.array(sitesF, dtype='int') # sites, folded into a conventional unit cell
        self.sites = np.array(sitesO, dtype='int') # sites, folded into a conventional unit cell
        self.clus = np.array(clus, dtype='int') # cluster associated with each site
        self.N = self.sites.shape[0]
        # print('liste of sites and folded sites:')
        # for i in range(self.N):
        #     print(i+1, '\t', self.sites[i], '\t', self.sitesF[i], '\tdiff = ', self.sitesF[i]-self.sites[i])
        # dictionary of sites (for getting the index from the position)
        for i,c in enumerate(sitesF):
            S = '[{:d},{:d},{:d}]'.format(c[0], c[1], c[2])
            self.siteF_index[S] = i



    #-----------------------------------------------------------------------------------------------
    def fold_shifted(self, r):
        """
        folds an integer vector back to the shifted unit cell of the superlattice
        This shifted unit cell is such that all of its vectors have positive or zero components in terms of superlattice vectors

        :param [int] r: integer component vector to be folded into the unit cell
        :returns: I, S, R where r = S + R and R is a superlattice vector, I is the index of the folded site
        :rtype: tuple[int, list[int], list[int]]

        """
        r = np.array(r, dtype=int)
        R = np.empty(3, dtype=int)
        S = np.empty(3, dtype=int)
        Q = np.empty(3, dtype=int)

        R = r@self.minors
        assert np.linalg.norm(R@self.E  - r*self.vol) < 1e-6, 'Folding error here : r = {}, R={}, check={}'.format(r,R, self.vol*r@np.linalg.inv(self.E))

        # now r = (R[0] e[0] + R[1] e[1] + R[2] e[2]) + (Q[0] e[0] + Q[1] e[1] + Q[2] e[2])/self.vol
        Q[0] = R[0] % self.vol
        R[0] //= self.vol
        if(Q[0] < 0):
            Q[0] += self.vol
            R[0] -= 1

        Q[1] = R[1] % self.vol
        R[1] //= self.vol
        if(Q[1] < 0):
            Q[1] += self.vol
            R[1] -= 1

        Q[2] = R[2] % self.vol
        R[2] //= self.vol
        if(Q[2] < 0):
            Q[2] += self.vol
            R[2] -= 1
	
        assert np.linalg.norm(R@self.E + Q@self.E//self.vol - r) < 1e-6, 'Folding error here : r = {}, R={}, Q={}'.format(r,R,Q)

        # S = self.E[0]*Q[0] + self.E[1]*Q[1] + self.E[2]*Q[2]
        S = np.dot(Q,self.E)
        S[0] //= self.vol
        S[1] //= self.vol
        S[2] //= self.vol
        Q = R
        # R = self.E[0]*Q[0] + self.E[1]*Q[1] + self.E[2]*Q[2]
        R = np.dot(Q,self.E)

  
        # then compensate between R and S to make the folding d-dimensional if d<3
        if(self.dim < 3):
            R -= self.E[2]*Q[2]
            S += self.E[2]*Q[2]
        if(self.dim  < 2):
            R -= self.E[1]*Q[1]
            S += self.E[1]*Q[1]

        Sstr = '[{:d},{:d},{:d}]'.format(S[0], S[1], S[2])
        if Sstr in self.siteF_index:
            I = self.siteF_index[Sstr]
        else:
            I = None

        assert np.linalg.norm(S+R-r)<1e-6, 'folding operation failed! r = {}, S={}, R={}'.format(r,S,R)
        return I, S, R

    #-----------------------------------------------------------------------------------------------
    def fold(self, r):
        """
        like 'fold_shifted', but this time folds back to the cluster itself, i.e., a different unit cell
        :param [int] r: integer component vector to be folded into the unit cell
        :returns: I, S, R where r = S + R and R is a superlattice vector, I is the index of the folded site
        :rtype: tuple[int, list[int], list[int]]
        """
        I, S, R = self.fold_shifted(r)
        if I != None:
            S += (self.sites[I] - self.sitesF[I])
            R += (self.sitesF[I] - self.sites[I])
        # print('fold : r = ', r, ' I = ', I, '  S = ', S, '  R = ', R)
        return I, S, R

    #-----------------------------------------------------------------------------------------------
    def draw(self, basis=None, plt_ax=None, labels=False):
        """
        Plots the sites, the shifted sites and the superlattice vectors
        :param list[list[float]] basis: the real space, geometric basis
        :param matplotlib.axes.Axes plt_ax: optional matplotlib axis object; if None, uses the current axis
        :param bool labels: if True, annotates each site with its index

        """
        if self.dim != 2:
            print('the draw() function only works in two dimensions. Skipping draw()')
            return

        if plt_ax is None:
            ax = plt.gca()
        else:
            ax = plt_ax

        if basis is None:
            basis = np.eye(3)
        
        e = self.e
        e = e@basis
        ax.set_aspect(1)
        ax.plot([0, e[0,0]], [0, e[0,1]], 'r-', lw=0.5)
        ax.plot([0, e[1,0]], [0, e[1,1]], 'r-', lw=0.5)
        
        # S = self.sitesF
        # S = S@basis
        # ax.plot(S[:,0], S[:,1], 'go', ms=3)

        S = self.sites
        S = S@basis
        ax.plot(S[:,0], S[:,1], 'bo', ms=6)

        eps = 0.1
        if(labels==True):
            for i in range(S.shape[0]):
                plt.text(S[i,0]+eps, S[i,1]-eps, '{:d}'.format(i+1), fontsize=8, ha='left', va='top')


        ax.set_axis_off()

        if plt_ax is None:
            plt.show()

    #-----------------------------------------------------------------------------------------------
    def draw_cdw(self, cdw, basis=None, plt_ax=None, labels=False, neighbors=True):
        """
        Plots the amplitudes of a cdw
        :param [float] cdw: the cdw amplitudes (same order as the sites)
        :param list[list[float]] basis: the real space, geometric basis
        :param matplotlib.axes.Axes plt_ax: optional matplotlib axis object; if None, uses the current axis
        :param bool labels: if True, annotates each site with its index
        :param bool neighbors: if True, also plots the CDW pattern for neighboring supercells

        """
        if self.dim != 2:
            print('the draw() function only works in two dimensions. Skipping draw()')
            return

        if plt_ax is None:
            ax = plt.gca()
        else:
            ax = plt_ax

        if basis is None:
            basis = np.eye(3)
        
        e = self.e
        e = e@basis
        plt.sca(ax)
        ax.set_aspect(1)
        ax.plot([0, e[0,0]], [0, e[0,1]], 'r-', lw=0.5)
        ax.plot([0, e[1,0]], [0, e[1,1]], 'r-', lw=0.5)
        S = self.sites
        S = S@basis
        fac = 8/np.max(np.abs(cdw))
        if neighbors:
            for j1 in range(-1,2,1):
                for j2 in range(-1,2,1):
                    es = j1*e[0,:] + j2*e[1,:]
                    for i in range(S.shape[0]):
                        ss = es + S[i,:]
                        if cdw[i] > 0: ax.plot(ss[0], ss[1], 'bo', ms=fac*cdw[i])
                        else: ax.plot(ss[0], ss[1], 'ro', ms=-fac*cdw[i])
                        ax.plot(ss[0], ss[1], 'k.', ms=1)
        else:
            for i in range(S.shape[0]):
                ss = S[i,:]
                if cdw[i] > 0: ax.plot(ss[0], ss[1], 'bo', ms=fac*cdw[i])
                else: ax.plot(ss[0], ss[1], 'ro', ms=-fac*cdw[i])
                ax.plot(ss[0], ss[1], 'k.', ms=1)
        if(labels==True):
            eps = 0.1
            for i in range(S.shape[0]):
                plt.text(S[i,0]+eps, S[i,1]-eps, '{:d}'.format(i+1), fontsize=8, ha='left', va='top')
        if plt_ax is None:
                plt.show()


    #-----------------------------------------------------------------------------------------------
    def cdw_eigenstates(self, _V, plt_ax=None, basis=np.eye(3), file=None, silent=False, neighbors=True):
        """
        Computes the possible CDW states of the cluster, to be used with the Hartree approximation

        :param [([int], float)] _V: density-density interactions
        :param matplotlib.axes.Axes plt_ax: optional matplotlib axis object; if not None, plots the interaction links on the sites
        :param list[list[float]] basis: the real space, geometric basis (identity by default)
        :param str file: if not None, saves the figure to this file path
        :param bool silent: if True, suppresses all printed output and figure generation
        :param bool neighbors: if True, plots the CDW patterns including neighboring supercells
        :returns: the array of eigenvalues, the matrix of eigenvectors, and the inter-cluster interaction matrix
        :rtype: tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]

        """

        V = []
        E = 0.0
        Vic = np.zeros((self.N, self.N))
        Vc = np.zeros((self.N, self.N))
        S2 = self.sites@basis
        if plt_ax != None:
            plt_ax.set_aspect(1.0)
            plt_ax.scatter(S2[:,0], S2[:,1], [24]*S2.shape[0],color='gray')

        assert type(_V) == list
        for y in _V:
            assert type(y) == tuple

        for v in _V:
            V.append((np.array(v[0], dtype=int), v[1]))

        if silent == False: print('-'*100)
        for v in V:
            for i,x in enumerate(self.sites):  # loop over site 1
                j, S, R = self.fold(self.sites[i] + v[0])
                if j != None:
                    if self.clus[i] == self.clus[j] and np.linalg.norm(R) < 1e-6:
                        Vc[i,j] += v[1]
                    else:
                        Vic[i,j] += v[1]
                    if plt_ax != None and silent == False:
                        DX = v[0]@basis
                        plt_ax.plot([S2[i,0], S2[i,0]+DX[0]], [S2[i,1], S2[i,1]+DX[1]])
                j, S, R = self.fold(self.sites[i] - v[0])
                if j != None:
                    if self.clus[i] == self.clus[j] and np.linalg.norm(R) < 1e-6:
                        Vc[i,j] += v[1]
                    else:
                        Vic[i,j] += v[1]
                    if plt_ax != None and silent == False:
                        DX = v[0]@basis
                        plt_ax.plot([S2[i,0], S2[i,0]-DX[0]], [S2[i,1], S2[i,1]-DX[1]])

        w, v = np.linalg.eigh(Vic)
        w = np.round(w,10)
        v = np.round(v,10)
        # making sure the first nonzero component of the eigenvector is positive
        for i in range(w.shape[0]):
            for j in range(w.shape[0]):
                if np.abs(v[j,i]) > 1e-6 and v[j,i] < 0:
                    v[:,i] *= -1
                    break


        if silent == False:
            print('intra-cluster V matrix:\n',Vc)
            print('inter-cluster V matrix:\n',Vic)

            fig, ax = plt.subplots((self.N+1)//2, 2, sharex=True, sharey=True)
            fig.set_size_inches(6, 3*ax.shape[0])
            if self.N > 2:
                ax = np.reshape(ax, 2*((self.N+1)//2))

            for i in range(self.N):
                print('\neigenvalue ', w[i], ' :\n', v[:, i])
                self.draw_cdw(v[:, i], basis=basis, plt_ax = ax[i], neighbors=neighbors)
                ax[i].set_title(r'No ${:d}, \lambda = {:f}$'.format(i, w[i]), fontsize=9)
                ax[i].tick_params(top=False, bottom=False, left=False, right=False, labelleft=False, labelbottom=False)

            # plt.tight_layout()
            if file != None: plt.savefig(file)
            else: plt.show()

        return w, v, Vic, Vc


    #-----------------------------------------------------------------------------------------------
    def cdw_energy(self, U, _V, n, cluster=False, pr=False):
        """
        Computes the potential energy per site associated with a given CDW pattern

        :param float U: on-site Hubbard interaction strength
        :param [([int], float)] V: density-density interactions
        :param [float] n: density pattern
        :param bool cluster: If True, limits the computation to the cluster (no inter-cluster interactions)
        :param bool pr: If True, prints progress
        :return float: potential energy

        """
        
        V = []
        E = 0.0
        n = np.array(n)
        assert type(_V) == list
        for y in _V:
            assert type(y) == tuple

        for v in _V:
            V.append((np.array(v[0], dtype=int), v[1]))

        # contribution of the on-site energy
        for i in range(self.N):
            E += 0.25*U*n[i]*n[i]

        if pr:
            print('\nComputing the CDW energy with n = ', n)
        for v in V:
            if cluster:
                for i,x in enumerate(self.sites):
                    R = x + v[0]
                    I = None
                    for j,r in enumerate(self.sites):
                        if np.linalg.norm(R-r) < 1e-6:
                            E += n[i]*n[j]*v[1]
                            break

            else:
                for i,x in enumerate(self.sitesF):
                    j, S, R = self.fold(x + v[0])
                    if j != None:
                        ener = n[i]*n[j]*v[1]
                        if np.abs(ener) > 1e-6 and pr:
                            print(v, '(', i+1, ',', j+1, ') --> ', ener)
                        E += ener
                
        return E/self.N
    
    #-----------------------------------------------------------------------------------------------
    def draw_V(self, _V, plt_ax, basis=None):
        """
        draws the links associated with extended interactions

        :param [([int], float)] _V: density-density interactions
        :param matplotlib.axes.Axes plt_ax: matplotlib axis object for plotting
        :param list[list[float]] basis: the real space, geometric basis
        :returns:None

        """
        
        V = []
        assert type(_V) == list
        for y in _V:
            assert type(y) == tuple

        for v in _V:
            V.append((np.array(v[0], dtype=int), v[1]))

        self.draw(basis, plt_ax)
        S2 = self.sites@basis
        color=['r', 'g', 'b', 'y', 'k']
        ic = 0
        for v in V:
            for i,x in enumerate(self.sitesF):
                j, S, R = self.fold_shifted(x + v[0])
                if j != None:
                    plt_ax.plot([S2[i,0],S2[j,0]], [S2[i,1],S2[j,1]], '-', c = color[ic%5])
            ic += 1
        plt.show()

    #-----------------------------------------------------------------------------------------------
    def draw_Vic(self, _V, plt_ax, basis=None):
        """
        draws the links associated with extended interactions

        :param [([int], float)] _V: density-density interactions
        :param matplotlib.axes.Axes plt_ax: matplotlib axis object for plotting
        :param list[list[float]] basis: the real space, geometric basis
        :returns:None

        """
        
        V = []
        assert type(_V) == list
        for y in _V:
            assert type(y) == tuple

        for v in _V:
            V.append((np.array(v[0], dtype=int), v[1]))

        self.draw(basis, plt_ax)
        S2 = self.sites@basis
        color=['r', 'g', 'b', 'y', 'k']
        ic = 0
        for v in V:
            DX = v[0]@basis
            for i,x in enumerate(self.sitesF):
                j, S, R = self.fold_shifted(x + v[0])
                
                if j != None:
                    plt_ax.plot([S2[i,0],S2[i,0]+DX[0]], [S2[i,1],S2[i,1]+DX[1]], '-', c = color[ic%5])
            ic += 1
        plt.show()

    #-----------------------------------------------------------------------------------------------
    def draw_mode(self, X, plt_ax, basis=None):
        """
        draws the links associated with extended interactions

        :param [float] X: amplitude (one per site)
        :param matplotlib.axes.Axes plt_ax: matplotlib axis object for plotting
        :param list[list[float]] basis: the real space, geometric basis
        :returns:None

        """
        
        S2 = self.sites@basis
        plt_ax.scatter(S2[:,0], S2[:,1], [24]*S2.shape[0],color='gray')
        X = X/np.linalg.norm(X)
        X *= 48
        col = ['b']*len(X)
        for i in range(len(X)):
            if X[i] < 0 :
                col[i] = 'r'
        X = np.abs(X)
        plt_ax.scatter(S2[:,0], S2[:,1], X, c=col)

    #-----------------------------------------------------------------------------------------------
    def sdw_energy(self, _J, sdw, cluster=False, pr=False):
        """
        Computes the energy per site associated with a given SDW pattern

        :param [([int], list[list[float]])] J: spin interactions
        :param list[list[float]] sdw: spin pattern (3 x N) : for each site, 3 angles are specifies : chi,theta,phi. The amplitude of the spin is sin(chi), its polar angle is theta and phi is the azimutal angle
        :param bool cluster: If True, limits the computation to the cluster (no inter-cluster interactions)
        :param bool pr: If True, prints progress
        :return float: energy

        """
        
        J = []
        E = 0.0
        assert type(_J) == list
        for y in _J:
            assert type(y) == tuple

        for v in _J:
            J.append((np.array(v[0], dtype=int), np.array(v[1])))

        sdw = np.array(sdw)
        if pr:
            print('\nComputing the SDW energy with sdw = ', sdw)
        s = np.zeros((self.N,3))
        for i in range(self.N):
            s[i,0] = np.abs(np.sin(sdw[0,i]))*np.sin(sdw[1,i])*np.cos(sdw[2,i])
            s[i,1] = np.abs(np.sin(sdw[0,i]))*np.sin(sdw[1,i])*np.sin(sdw[2,i])
            s[i,2] = np.abs(np.sin(sdw[0,i]))*np.cos(sdw[1,i])
        for v in J:
            if cluster:
                for i,x in enumerate(self.sites):
                    R = x + v[0]
                    for j,r in enumerate(self.sites):
                        if np.linalg.norm(R-r) < 1e-6:
                            E += s[i]@v[1]@s[j]
                            break

            else:
                for i,x in enumerate(self.sitesF):
                    j, S, R = self.fold(x + v[0])
                    if j != None:
                        ener = s[i]@v[1]@s[j]
                        # if np.abs(ener) > 1e-6 and pr:
                        #     print(v, '(', i+1, ',', j+1, ') --> ', ener)
                        E += ener
                
        return E/self.N

    #-----------------------------------------------------------------------------------------------
    def convert_to_spins(self, x, pr=False):
        """
        Converts a flat array of angular coordinates into a spin vector array.
        Each site is described by three angles: chi (amplitude), theta (polar), phi (azimuthal).

        :param [float] x: flat array of 3*N angular values (chi, theta, phi for each site), reshaped to (3, N)
        :param bool pr: if True, prints each site position and its spin vector
        :returns: spin vectors for each site, shape (N, 3)
        :rtype: numpy.ndarray
        """
        s = np.zeros((self.N,3))
        sdw = np.reshape(x,(3,8))
        for i in range(self.N):
            s[i,0] = np.abs(np.sin(sdw[0,i]))*np.sin(sdw[1,i])*np.cos(sdw[2,i])
            s[i,1] = np.abs(np.sin(sdw[0,i]))*np.sin(sdw[1,i])*np.sin(sdw[2,i])
            s[i,2] = np.abs(np.sin(sdw[0,i]))*np.cos(sdw[1,i])

            if pr:
                print(self.sites[i], '\t', s[i,:])
        return s



import pyqcm
from pyqcm.cdmft import *

ns = 4
nb = 4
no = ns + nb
CM = pyqcm.cluster_model( ns, nb)  # call to constructor of class "cluster_model"

CM.new_operator('eb1', 'one-body', ((ns+1,ns+1,1), (ns+2,ns+2,1), (ns+no+1,ns+no+1,1), (ns+no+2,ns+no+2,1)))
CM.new_operator('eb2', 'one-body', ((ns+3,ns+3,1), (ns+4,ns+4,1), (ns+no+3,ns+no+3,1), (ns+no+4,ns+no+4,1)))
CM.new_operator('tb1', 'one-body', ((1,ns+1,1), (1+no,ns+no+1,1), (4,ns+2,1), (4+no,ns+no+2,1)))
CM.new_operator('tb2', 'one-body', ((1,ns+3,1), (1+no,ns+no+3,1), (4,ns+4,1), (4+no,ns+no+4,1)))

clus = pyqcm.cluster(CM, [(i,0,0) for i in range(ns)]) # call to constructor of class "cluster"

model = pyqcm.lattice_model('1D_4_4b', clus, ((ns,0,0),)) # call to constructor of class "lattice_model"

model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)

############# model definition end #################

model.set_target_sectors(['R0:N8:S0'])
model.set_parameters("""
U = 4
mu = 2
t = 1
eb1_1 = 1
eb2_1 = -1
tb1_1 = 0.5
tb2_1 = 0.5
""")

I = pyqcm.model_instance(model) # call to constructor of class "model_instance"
I.spectral_function(path='line', file='test_reset1.pdf')

cdmft = CDMFT(model, ('eb1_1', 'eb2_1', 'tb1_1', 'tb2_1'))

############# new model definition #################
pyqcm.reset_model()

ns = 2
nb = 4
no = ns + nb
CM = pyqcm.cluster_model( ns, nb)  # call to constructor of class "cluster_model"

CM.new_operator('eb1', 'one-body', ((ns+1,ns+1,1), (ns+2,ns+2,1), (ns+no+1,ns+no+1,1), (ns+no+2,ns+no+2,1)))
CM.new_operator('eb2', 'one-body', ((ns+3,ns+3,1), (ns+4,ns+4,1), (ns+no+3,ns+no+3,1), (ns+no+4,ns+no+4,1)))
CM.new_operator('tb1', 'one-body', ((1,ns+1,1), (1+no,ns+no+1,1), (2,ns+2,1), (2+no,ns+no+2,1)))
CM.new_operator('tb2', 'one-body', ((1,ns+3,1), (1+no,ns+no+3,1), (2,ns+4,1), (2+no,ns+no+4,1)))

clus = pyqcm.cluster(CM, [(i,0,0) for i in range(ns)]) # call to constructor of class "cluster"

model = pyqcm.lattice_model('1D_2_4b', clus, ((ns,0,0),)) # call to constructor of class "lattice_model"

model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)

model.set_target_sectors(['R0:N6:S0'])
model.set_parameters("""
U = 4
mu = 2
t = 1
eb1_1 = 1
eb2_1 = -1
tb1_1 = 0.5
tb2_1 = 0.5
""")

I = pyqcm.model_instance(model) # call to constructor of class "model_instance"
I.spectral_function(path='line', file='test_reset2.pdf')

cdmft = CDMFT(model, ('eb1_1', 'eb2_1', 'tb1_1', 'tb2_1'))



import pyqcm
from pyqcm.hartree import *
from pyqcm.loop import *
from pyqcm.vca import *
import model_1D_4

pyqcm.set_target_sectors(['R0:N4:S0'])
pyqcm.set_parameters("""
t=1
U=4
V=4
mu=10
V0m = 5
# V1m = -0.15
# cdw=0
# cdw_1=0.09
""")

V0m_H = hartree('V0m', 'V', 1.0, accur=1e-4, lattice=True)
# V1m_H = hartree('V1m', 'V', 1.0, accur=1e-4, lattice=True)
# plot_sef('V1m', np.arange(-0.5,0.5,0.02), hartree=[V1m_H]); exit()


def F():
    pyqcm.new_model_instance()
    pyqcm.averages()

Hartree(F, [V0m_H], maxiter=10, eps_algo=0); exit()

first = True
for v in np.arange(4, 3, -0.02):
    pyqcm.set_parameter('V', v)
    pyqcm.set_parameter('mu', 2+2*v)
    # Hartree(F, [V0m_H, V1m_H], maxiter=10, eps_algo=0)
    # vca(varia=['cdw_1', 'V0m', 'V1m'], steps=[0.01]*3, accur=[1e-4, 1e-2, 1e-2], max=[10]*3, hartree=[V0m_H, V1m_H], max_iter=100)
    # vca(varia=['cdw_1', 'V0m'], steps=[0.01]*2, accur=[1e-4, 1e-2], max=[10]*2, hartree=[V0m_H], max_iter=100)
    vca(varia=['cdw_1'], steps=[0.01], accur=[1e-4], max=[10], hartree=[V0m_H], max_iter=100, hartree_self_consistent=True)
    # f = open('cdw.tsv', 'a')
    # des, val = pyqcm.properties()
    # if first:
    #     f.write(des + '\n')
    #     first = False
    # f.write(val + '\n')
    # f.close()





####################################################################################################
"""
VCA for antiferromagnetism
Find the VCA solution for Néel antiferromagnetism in the nearest-neighbor, particle-hole symmetric Hubbard model at half-filling, using **two** $3\times 3$ clusters embedded in a two-dimensional square lattice.
Use only one variational parameter: the Weiss field $M$, defined as the coefficient of the antiferromagnetic operator. Plot the order parameter $\langle M\rangle$ as a function of $U$, from $U=10$ to $U=0$, in steps of $\Delta U=0.5$.
"""
####################################################################################################
import pyqcm
import pyqcm.vca as vca
import numpy as np

pyqcm.set_global_parameter('accur_SEF', 1e-4)

# declare a cluster model of 9 sites, named 'clus'
CM = pyqcm.cluster_model(9)  

sites = []
for j in range(3):
    for i in range(3):
        sites.append((i,j,0))

# defining a hopping operator that only hops around the perimeter of the clusters
ns = 9
CM.new_operator('tperim', 'one-body',[
    (1, 2, -1.0), 
    (2, 3, -1.0),
    (3, 6, -1.0), 
    (6, 9, -1.0),
    (8, 9, -1.0), 
    (7, 8, -1.0),
    (4, 7, -1.0), 
    (1, 4, -1.0),
    (1+ns, 2+ns, -1.0), 
    (2+ns, 3+ns, -1.0),
    (3+ns, 6+ns, -1.0), 
    (6+ns, 9+ns, -1.0),
    (8+ns, 9+ns, -1.0), 
    (7+ns, 8+ns, -1.0),
    (4+ns, 7+ns, -1.0), 
    (1+ns, 4+ns, -1.0)
])

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus1 = pyqcm.cluster(CM, sites) 
clus2 = pyqcm.cluster(CM, sites, (3,0,0)) # second cluster, offset from the first

# define a lattice model named '1D_4' made of the cluster(s) clus and superlattice vector (4,0,0)
model = pyqcm.lattice_model('model_3x3_2C', (clus1, clus2), ((6,0,0),(1,3,0)))

# define a few operators in this model
model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)
model.hopping_operator('t', (0,1,0), -1)
model.density_wave('M', 'Z', (1,1,0)) # Spin density wave at Q = (pi,pi)

# model.draw_operator('U', values=True); exit()

# model.draw_operator('t')
# model.draw_cluster_operator(model.clus[0], 'tperim')

# Setting target sectors with S=1 and S=-1 such that S_total = 0
model.set_target_sectors(['R0:N9:S-1/R0:N9:S1','R0:N9:S-1/R0:N9:S1'])
model.set_parameters("""
t = 1
U = 8
mu = 0.5*U
M = 0
M_1 = 0.1
M_2 = 1*M_1
# t_1 = 1
# t_2 = 1*t_1
# tperim_1 = 1e-9
# tperim_2 = 1*tperim_1
""")


#---------------------------------------------------------------------------------------------------
# Estimating the starting value of `M_1` 
# Here we make a rough plot of the self-energy functional as a function of `M_1` to find a good starting point for the VCA procedure.

# model.set_parameter("U", 8)
I = pyqcm.model_instance(model)
print(I.ground_state())
# pyqcm.vca.plot_sef(model, 'M_1', np.arange(0.05, 0.151, 0.01)); exit()


#---------------------------------------------------------------------------------------------------
# Controlled VCA loop over $U$
# Here we loop over U in the range from 10 to 0 and perform the VCA with M_1 as a variational parameter. 
# Looping downwards over U allows for easier convergence on a solution
# We use the function controlled_loop() to benefit from predictors.

model.set_parameter("M_1", 0.11) # initial guess for M_1 based on the sef
I = pyqcm.model_instance(model)

def run_vca():
    V = vca.VCA(model, varia='M_1', steps=0.005, accur=2e-3, max=10, method='altNR')
    return V.I

# controlled_loop performs successive VCAs with starting value prediction
model.controlled_loop(run_vca, varia=["M_1"], loop_param="U", loop_range=(10, -0.1, -0.5)) # loops from 10 to 0 in increments of 0.5


#---------------------------------------------------------------------------------------------------
# VCA for U=8 with M_1 and t_1 as variational parameters

model.set_parameter("U", 8)
V = vca.VCA(model, varia=('M_1', 't_1'), start=(0.17, 1.15), steps=(0.005, 0.005), accur=(2e-3, 2e-3), max=(10, 10)) 


#---------------------------------------------------------------------------------------------------
# VCA for U=8 with M_1, t_1 and tperim_1 as variational parameters

V = vca.VCA(model, varia=('M_1', 't_1', 'tperim_1'), start=(0.17, 1.15, 0.04), steps=(0.005, 0.005, 0.005), accur=(2e-3, 2e-3, 2e-3), max=(10, 10, 10))

exit()
#---------------------------------------------------------------------------------------------------
# Getting data from prerun VCAs
d_M = np.genfromtxt("./example_data/EX2_data_M.tsv", names=True)
d_M_t = np.genfromtxt("./example_data/EX2_data_M_t.tsv", names=True)
d_M_t_tperim = np.genfromtxt("./example_data/EX2_data_M_t_tperim.tsv", names=True)


# #### Plotting $|\langle M \rangle|$ as a function of $U$ for all three simulations

# In[ ]:


fig, ax = plt.subplots()

ax.plot(d_M["U"], np.abs(d_M["ave_M"]), "s", markersize=3, label="M_1")
ax.plot(d_M_t["U"], np.abs(d_M_t["ave_M"]), "s", markersize=3, label="M_1 & t_1")
ax.plot(d_M_t_tperim["U"], np.abs(d_M_t_tperim["ave_M"]), "s", markersize=3, label="M_1, t_1 & tperim_1")

ax.legend(loc="upper left")

ax.set_xlabel("U")
ax.set_ylabel("$|\langle M\\rangle|$")

subax = fig.add_axes([0.5, 0.25, 0.35, 0.35])

subax.plot(d_M["U"][14:17], np.abs(d_M["ave_M"][14:17]), "s", markersize=3)
subax.plot(d_M_t["U"], np.abs(d_M_t["ave_M"]), "s", markersize=3)
subax.plot(d_M_t_tperim["U"], np.abs(d_M_t_tperim["ave_M"]), "s", markersize=3)

subax.set_xlabel("U")
subax.set_ylabel("$|\langle M\\rangle|$")

subax.set_xticks((d_M["U"][14:17]))

fig.show()


# #### Plotting the Weiss field as a function of $U$ for all three simulations

# In[ ]:


fig, ax = plt.subplots()

ax.plot(d_M["U"], d_M["M_1"], "s", markersize=3, label="M_1")
ax.plot(d_M_t["U"], d_M_t["M_1"], "s", markersize=3, label="M_1 & t_1")
ax.plot(d_M_t_tperim["U"], d_M_t_tperim["M_1"], "s", markersize=3, label="M_1, t_1 & tperim_1")

ax.legend(loc="lower right")

ax.set_xlabel("U")
ax.set_ylabel("|M_1|")

fig.show()


# ### Interpretation
# 
# In the first figure, we can see how $|\langle M \rangle|$ seems to approach 1 relatively quickly as a function of $U$. On the other hand, $|M\_1|$ seems to peak around $U\approx 5$ (fig.2). This means that as $U$ rises above 5, a rise in order parameter actually requires a *diminishing* Weiss field due to the rising interaction strength.
# 
# In both figures, adding in t_1 to the variational parameters seems to cause a rise in the values. However, adding in tperim_1 only causes a very minute increase. This makes sense since there doesn't seem to be a good reason for hopping exclusively around the perimeter of the cluster as described in the definition of the given operator.



import pyqcm
pyqcm.set_global_parameter('nosym')

L=4
from model_1D import model1D
model = model1D(4, sym=False)

mixing = 0
cluster = False

if mixing == 4:
    model.set_target_sectors(['R0:N{:d}:S0'.format(L)])
    model.set_parameters("""
    U = 4
    t=1
    tsi=1
    mu = 1
    """)

elif mixing == 3:
    model.set_target_sectors(['R0'])
    model.set_parameters("""
    U = 4
    t=1
    tsi=1
    S = 1e-9
    Hx = 1e-9
    mu = 1
    """)

elif mixing == 2:
    model.set_target_sectors(['R0:N{:d}'.format(L)])
    model.set_parameters("""
    U = 4
    t=1
    tsi=1
    H = -1
    Hx = 1e-8 
    mu = 1
    """)

elif mixing == 1:
    model.set_target_sectors(['R0:S0'])
    model.set_parameters("""
    U = 4
    t=1
    tsi=1
    S = 1e-9
    mu = 1
    """)

elif mixing == 0:
    model.set_target_sectors(['R0:N{:d}:S0'.format(L)])
    model.set_parameters("""
U = 4
# t=1
ti=1
mu = 1
    """)

import matplotlib.pyplot as plt

# plt.gcf().set_size_inches(12/2.54, 12/2.54)

I = pyqcm.model_instance(model)
if cluster:
    I.cluster_spectral_function(wmax = 6, plt_ax = plt.gca(), full=True, offset=14)
else:
    I.spectral_function(wmax = 6, plt_ax = plt.gca(), path='line')

I = pyqcm.model_instance(model)
pyqcm.set_global_parameter('continued_fraction')
I = pyqcm.model_instance(model)
if cluster:
    I.cluster_spectral_function(wmax = 6, full=True, offset=14, plt_ax = plt.gca(), color='r')
else:
    I.spectral_function(wmax = 6, plt_ax = plt.gca(), path='line', color='r')

plt.savefig('test_CF.pdf')


import numpy as np
import pyqcm
import matplotlib.pyplot as plt
import model_2x2_C2 as M

I = pyqcm.model_instance(M.model)

ax = None
F = None
# F = None
#-----------------------------------------------------------------
# import record_2x2_anom
# import record_spin

# I.plot_dispersion(); exit()
##################################################################
# TEST UNITAIRE

K = np.array([(1,1,0.)])
print('testing dispersion()')
print(I.dispersion(K))

print('testing epsilon()')
print(I.epsilon(K))



# """
# BATTERY OF TESTS TO VERIFY THE INTEGRITY AND FUNCTIONALITY OF PYQCM
# """

import unittest
import os

TEST_FILES = []
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__)) # gets current path of script

if not os.path.exists(f"{SCRIPT_DIR}/test_outputs"): # checks if the output dir exists and if not creates it
    os.mkdir(f"{SCRIPT_DIR}/test_outputs")

os.chdir(f"{SCRIPT_DIR}/test_outputs") # changes working dir to the test output dir

for file in os.listdir(f"{SCRIPT_DIR}/test_files"): # finds all test files in the testing directory following the pattern test***********.py
    if file[:4]=="test" and file[-3:]==".py":
        TEST_FILES.append(file)

def run_file(test_name):
    """Runs a file of a given name inside the testing directory. Mostly aesthetic.
    """
    return os.system(f"python3 {SCRIPT_DIR}/test_files/{test_name}")


class TestAll(unittest.TestCase):

    # generates the appropriate methods given the detected test files
    for file in TEST_FILES:
        exec(f"""def {file[:-3]}(self): \n\tself.assertFalse(run_file("{file}"))""")

if __name__ == "__main__":
    unittest.main()


from pyqcm import *

L = 8
new_cluster_model('clus', L, 0)
pos = []
for i in range(L):
    pos.append((i,0,0))


add_cluster('clus', [0, 0, 0], pos)
lattice_model('1D_{:d}'.format(L), [[L, 0, 0]])
interaction_operator('U')
interaction_operator('V', link=( 1, 0, 0))
hopping_operator('t', ( 1, 0, 0), -1)  # NN hopping
hopping_operator('ti', ( 1, 0, 0), -1, tau=2)  # NN hopping
hopping_operator('tp', ( 2, 0, 0), -1)  # NNN hopping
hopping_operator('hx', [0, 0, 0], 1, tau=0, sigma=1)  # field in the x direction
anomalous_operator('D', ( 1, 0, 0), 1)  # NN singlet



"""Looping utilities for VCA and CDMFT calculations.

Provides functions for iterating over parameter spaces with predictors, including
:func:`linear_loop`, :func:`controlled_loop`, :func:`fade`, :func:`Hartree_procedure`,
and file-driven loops. These functions are added as methods of the lattice model class.
"""
import numpy as np
import pyqcm
import os

first_time=True
####################################################################################################
# Internal functions

def _predict_bad(x, y, p):
	for i in range(len(x)):
		if abs(x[i] - y[i]) / (abs(x[i]) + 0.01) > p:
			return True
	return False


#---------------------------------------------------------------------------------------------------
def _predict_good(x, y, p):
	for i in range(len(x)):
		if abs(x[i] - y[i]) / abs((x[i]) + 0.01) > p:
			return False
	return True

#---------------------------------------------------------------------------------------------------
def _poly_predictor(X,Y,i,n):
	x = X[i-n-1:i]  # the n+1 previous values of X
	y = Y[i-n-1:i]  # the n+1 previous values of Y
	pol = np.polyfit(x, y, n)
	return np.polyval(pol, X[i])



####################################################################################################
# Additional methods of the model class

#---------------------------------------------------------------------------------------------------
def loop_from_file(self, task, file):
	"""Performs a task 'task' over a set of model parameters defined in a file. The definition of each
	model instance must be done in the task 'task'; it is not done by this looping function.

	:param task: function (task) to perform
	:param str file: name of the data file specifying the solutions, in the same tab-separated format usually used to write solutions

	"""
	param = self.parameter_set()
	data = np.genfromtxt(file, names=True, dtype=None, encoding='utf8')

	if len(data.shape) == 0 : data = np.expand_dims(data, axis=0)

	print('number of data sets: ', len(data))

	independent_params = []
	for key, val in param.items():
		if val[1] is None and key in data.dtype.names:
			independent_params.append(key)

	print('independent parameters: ', independent_params)

	N = len(data)
	for i in range(N):
		print('*'*30, ' set ', i+1, ' of ', N, '*'*30)
		for x in independent_params:
			print(x, ' --> ', data[x][i])
			self.set_parameter(x, data[x][i])
		task()


#---------------------------------------------------------------------------------------------------
def loop_from_table(self, task, data):
	"""Performs a task 'task' over a set of model parameters defined in a table. The definition of each
	model instance must be done in the task 'task'; it is not done by this looping function.

	:param task: function (task) to perform
	:param ndarray data: table of parameters, with named columns, as read or filtered from numpy.genfromtxt()

	"""
	param = self.parameter_set()
	print('number of data sets: ', len(data))

	independent_params = []
	for key, val in param.items():
		if val[1] is None and key in data.dtype.names:
			independent_params.append(key)

	print('independent parameters: ', independent_params)

	N = len(data)
	for i in range(N):
		print('*'*30, ' set ', i+1, ' of ', N, '*'*30)
		for x in independent_params:
			print(x, ' --> ', data[x][i])
			self.set_parameter(x, data[x][i])
		task()


#---------------------------------------------------------------------------------------------------
# performs a loop (a linear trajectory) between two subsets of parameters
def linear_loop(self, N, task, varia=None, params = None, predict=True):
	"""
	Performs a loop with a predictor. The definition of each
	model instance must be done in 'task'; it is not done by this looping function.

	:param N: number of intervals within the loop
	:param task: function called at each step of the loop
	:param [str] varia: names of the variational parameters
	:param dict params: dict mapping parameter names to 2-tuples of (initial, final) values
	:param bool predict: if True, uses a linear or quadratic predictor
	:returns: None

	"""

	pyqcm.banner('linear loop over {:d} values'.format(N), '%', skip=1)
	if type(params) != dict:
		raise TypeError('the argument "P" of linear_loop must be a dict of 2-tuples')
	if varia is None:
		raise pyqcm.MissingArgError('variational parameters (varia) must be specified')
	if type(varia) != list:
		raise TypeError('the argument "varia" of linear_loop must be a list')

	current_value = {}

	nvar = len(varia)  # number of variational parameters
	sol = np.empty((4, nvar))  # current solution + 3 previous solutions
	start = [0]*nvar
	for iter in range(N+1):
		# sets the control parameters of the loop
		for x in params:
			current_value[x] = params[x][0]*(N-iter)/N + params[x][1]*iter/N
			self.set_parameter(x, current_value[x])
		
		sol = np.roll(sol, 1, 0)  # cycles the solutions
		# predicting the starting value from previous solutions (do nothing if loop_counter=0)
		fit_type = ''
		if iter == 1 or not predict:
			start = sol[1, :]  # the starting point is the previous value in the loop
		elif iter > 1:
			if iter == 2:
				fit_type = ' (linear fit)'
				x = [iter-1, iter-2]  # the two previous values of the loop parameter
				y = sol[1:3, :]  # the two previous values of the loop parameter
				pol = np.polyfit(x, y, 1)
			else:
				fit_type = ' (quadratic fit)'
				x = [iter-1, iter-2, iter-3]  # the three previous values of the loop parameter
				y = sol[1:4, :]  # the three previous values of the loop parameter
				pol = np.polyfit(x, y, 2)
			for i in range(nvar):
				start[i] = np.polyval(pol[:, i], iter)

		pyqcm.banner('loop index = {:d}'.format(iter + 1), '=')
		for x in params:
			print(x, ' ===> {:1.4f}'.format(current_value[x]))
		if iter > 0:
			print('predictor: ', start, fit_type)
		task()
		sol[0,:] = self.parameters(varia)

	pyqcm.banner('linear loop ended normally', '%')

#---------------------------------------------------------------------------------------------------
# performs a loop for VCA or CDMFT with an external loop parameter and an optional control on an order parameter

def controlled_loop(self, task, varia=None, loop_param=None, loop_range=None, control_func=None, retry=None, max_retry=4, predict=True):
	"""
	Performs a controlled loop for VCA or CDMFT with a predictor.  The definition of each
	model instance must be done and returned by 'task'; it is not done by this looping function.

	:param task: a function called at each step of the loop. Must return a model_instance.
	:param [str] varia: names of the variational parameters
	:param str loop_param: name of the parameter looped over
	:param (float, float, float) loop_range: range of the loop (start, end, step)
	:param control_func: (optional) name of the function that controls the loop (returns bool). Takes a model instance as argument
	:param char retry: If None, stops on failure. If 'refine', adjusts the step (divide by 2) and retry. If 'skip', skip to next value.
	:param int max_retry: Maximum number of retries of either type
	:param bool predict: if True, uses a linear or quadratic predictor

	"""
	if retry not in (None, 'skip', 'refine'): 
		raise ValueError('"retry" argument must be None, "refine" or "skip".')

	pyqcm.banner('controlled loop over '+loop_param, '%', skip=1)
	if loop_param is None:
		raise pyqcm.MissingArgError('The name of the control parameter (loop_param) is missing')
	if loop_range is None:
		raise pyqcm.MissingArgError('A loop range (min, max, step) must be provided')
	if (loop_range[1]-loop_range[0])*loop_range[2] < 0:
		raise ValueError('the loop range is incoherent: step has wrong sign')
	if varia is None:
		raise pyqcm.MissingArgError('variational parameters (varia) must be specified')
	if pyqcm.is_sequence(varia) == False:
		varia = (varia,)

	nvar = len(varia)  # number of variational parameters
	sol = np.empty((4, nvar))  # current solution + 3 previous solutions
	prm = np.empty(4)  # current loop parameter + 3 previous values

	# prm[0] = loop_range[0] - loop_range[2]
	prm[0] = loop_range[0]
	step = loop_range[2]
	loop_counter = 0
	try_again = False
	nretry = 0
	fac = 1
	start = [0]*nvar
	abort = False
	while True:  # main loop here
		self.set_parameter(loop_param, prm[0])
		pyqcm.banner('loop index = {:d}, {:s} = {:1.4f}'.format(loop_counter + 1, loop_param, prm[0]), '=', skip=1)
		try:
			I = task()

		except Exception as E:
			if loop_counter == 0 or not retry:
				raise ValueError('Out of bound on starting values in controlled_loop(), aborting') from E
			else:
				try_again = True
			
		if control_func is not None:
			if not control_func(I):
				if loop_counter < 2 and retry == None:
					print('Control failed on starting. Maybe try with different starting point')
					break
				elif retry == None:
					print('Control failed. Aborting loop.')
					abort = True
					break
				try_again = True
	
		if try_again:
			if retry == 'refine':
				step = step / 2
				pyqcm.banner('retrying with half the step', '=')
			elif retry == 'skip':
				fac += 1
				pyqcm.banner('skipping to next value', '=')
			try_again = False
			nretry += 1
			if nretry > max_retry:
				break
		else:
			prm = np.roll(prm, 1)  # cycles the values of the loop parameter
			sol = np.roll(sol, 1, 0)  # cycles the solutions
			nretry = 0
			fac = 1

		prm[0] = prm[1] + fac*step  # updates the loop parameter

		sol[0,:] = I.parameters(varia)

		if loop_range[1] >= loop_range[0] and prm[0] > loop_range[1]:
			break
		if loop_range[1] <= loop_range[0] and prm[0] < loop_range[1]:
			break
	
		# predicting the starting value from previous solutions (do nothing if loop_counter=0)
		fit_type = ''
		if loop_counter == 1 or not predict:
			start = sol[1, :]  # the starting point is the previous value in the loop
		elif loop_counter > 1:
			if loop_counter == 2:
				fit_type = ' (linear fit)'
				x = prm[1:3]  # the two previous values of the loop parameter
				y = sol[1:3, :]  # the two previous values of the loop parameter
				pol = np.polyfit(x, y, 1)
			else:
				fit_type = ' (quadratic fit)'
				x = prm[1:4]  # the three previous values of the loop parameter
				y = sol[1:4, :]  # the three previous values of the loop parameter
				pol = np.polyfit(x, y, 2)
			for i in range(nvar):
				start[i] = np.polyval(pol[:, i], prm[0])

		if loop_counter > 0:
			print('predictor : ', start, fit_type)
			for i in range(len(varia)):
				self.set_parameter(varia[i], start[i])
				print(' ---> ', varia[i], ' = ', start[i])

		if loop_counter > 2 and not retry and retry:
			if(_predict_good(sol[0, :], start, 0.01) and step <  loop_range[2]):
				step *= 1.5
				print('readjusting step to ', step)
			elif(_predict_bad(sol[0, :], start, 0.2) and step > loop_range[2] / 10):
				step /= 1.5
				print('readjusting step to ', step)

		loop_counter += 1

	if nretry > max_retry:
		pyqcm.banner('controlled loop ended on signal by control function', '%')
		return 1
	elif abort:
		pyqcm.banner('controlled loop ended with error', '%')
		return 2
	else:
		pyqcm.banner('controlled loop ended normally', '%')
		return 0


#---------------------------------------------------------------------------------------------------
def fade(self, task, P, n):
	"""
	Fades the model between two sets of parameters, in n steps

	:param task: task to perform within the loop
	:param dict P: dict of parameters with a tuple of values (start and finish) for each
	:param n: number of steps
	:returns: None

	"""

	lambda_array = np.linspace(0.0, 1.0, n)
	for L in lambda_array:
		for p in P:
			self.set_parameter(p, L*P[p][1] + (1-L)*P[p][0])
		task()

#---------------------------------------------------------------------------------------------------
def controlled_fade(self, task, P, n, C, file='fade.tsv', tol=1e-4, method='Broyden', maxiter=32, alpha=0.0, eps_algo=0, bracket=None, delta=None, verb=False):
# IN DEVELOPMENT
	"""
	Fades the model between two sets of parameters, in n steps

	:param task: task to perform within the loop
	:param dict P: dict of parameters with a tuple of values (start and finish) for each
	:param n: number of steps
	:param dict C: dict of adjustable parameters whose average (the value of C) should stay fixed during the fade
	:param str file: file to which the converged results are written
	:param float tol: precision on the averages
	:param str method: method used for solving the nonlinear constraints
	:param maxiter: maximum number of iterations for solving the constraints
	:param float alpha: parameter used in some non linear solvers (broyden)
	:param int eps_algo: convergence accelerator parameter in fixed_point
	:param (float,float) bracket: bracket used in some nonlinear 1D solvers
	:param float delta: interval used to update the bracket used in some nonlinear 1D solvers
	:param bool verb: If True, writes progress reports

	"""
	assert type(P) is dict
	assert type(C) is dict
	for x in P:
		assert type(P[x]) is tuple

	print('Fading procedure in ', n, 'steps :')
	for x in P:
		print(x, ' from ', P[x][0], ' to ', P[x][1])

	# A = np.zeros(len(C)) # allocating the array of fixed average values 
	A = np.array([C[y] for y in C])
	nconstr = A.shape[0]
	
	def F(x):
		global I_current
		for y in C :
			self.set_parameter(y, x, pr=True)
		I = task()
		ave = I.averages()
		B = np.array([ave[y] for y in C]) # average values
		if verb: print('-------> root search function value: ', B-A)
		return B-A

	par = self.parameters()
	Cv = np.array([par[x] for x in C])
	x0 = Cv
	Z = np.linspace(0.0, 1.0, n, endpoint=True)

	from scipy.optimize import root, root_scalar
	methods = ['hybr', 'lm', 'broyden1', 'broyden2', 'anderson', 'linearmixing', 'diagbroyden', 'excitingmixing', 'krylov', 'df-sane']

	methods1D = ['bisect','brentq','brenth','ridder','toms748','newton','secant','halley']

	for i, z in enumerate(Z):
		for p in P:
			self.set_parameter(p, z*P[p][1] + (1-z)*P[p][0], pr=True)
		try:
			alpha = X[2]
		except (IndexError, TypeError):
			alpha = 0.0
		if method == 'Broyden' : 
			X = pyqcm.broyden(F, x0, iJ0 = alpha, xtol=tol, maxiter=maxiter)
		elif method == 'fixed_point' : 
			X = pyqcm.fixed_point_iteration(F, x0, xtol=tol, maxiter=maxiter,  alpha = alpha, eps_algo=eps_algo)
		elif nconstr==1 and method in methods1D:
			sol = root_scalar(F, method=method, bracket=bracket, fprime=None, fprime2=None, x0=x0, x1=None, xtol=tol, maxiter=None, options=None)
			if sol.converged is False:
				raise ValueError('Failure of the scipy root_scalar method ' + method + ' in controlled_fade:\n' + sol.flag)
			X = (sol.root,)
			bracket = (sol.root-delta, sol.root+delta)
		elif method in methods:
			sol_root = root(F, x0, method=method, jac=None, tol=tol, callback=None)
			if sol_root.success is False:
				raise ValueError('Failure of the scipy root method ' + method + ' in controlled_fade:\n' + sol_root.message)
			X = (sol_root.x,)
		I = pyqcm.model_instance(self)
		I.averages()
		I.write_summary(file)
		x0 = X[0]
			

#---------------------------------------------------------------------------------------------------
def Hartree_procedure(self, task, couplings, maxiter=32, iteration='fixed_point', eps_algo=0, file='hartree.tsv', SEF=False, pr = False, alpha=0.0):
	"""
	Performs the Hartree approximation
	
	:param task: task to perform within the loop. Must return a model_instance
	:param [hartree] couplings: sequence of couplings (or single coupling)
	:param int maxiter: maximum number of iterations
    :param str iteration: method of iteration of parameters ('fixed_point' or 'broyden')
	:param int eps_algo: number of elements in the epsilon algorithm convergence accelerator = 2*eps_algo + 1 (0 = no acceleration)
	:param str file: name of the file to which the converged result is written via write_summary()
	:param bool SEF: if True, computes the Potthoff functional at the end of the procedure
	:param bool pr: if True, prints progress of each coupling update
    :param float alpha: if iteration='fixed_point', damping parameter, otherwise trial inverse Jacobian (if a float, that is (1+alpha)*unit matrix)
	:returns: model instance, converged inverse Jacobian

	"""
	
	global first_time

	if pyqcm.is_sequence(couplings) is False:
		couplings = (couplings,)

	pyqcm.banner('Hartree procedure', c='*', skip=1)
	hartree_converged = False
	diff_tot = 1e6
	var_data = np.empty((len(couplings), maxiter+2))

	n = len(couplings)
	P = self.parameters()
	X = np.empty(n)
	for i,C in enumerate(couplings):
		X[i] = P[C.Vm]


	def F(x):
		x_new = np.copy(x)
		for i,C in enumerate(couplings):
			self.set_parameter(C.Vm, x[i])
		I = task()
		for i,C in enumerate(couplings):
			C.update(I, pr=pr)
			C.print()
			x_new[i] = C.vm
		return x - x_new
	
	def G():
		hartree_converged = True
		for i,C in enumerate(couplings):
			hartree_converged = hartree_converged and C.converged()
		return hartree_converged
	
	niter = 0
	if iteration == 'broyden' or iteration == 'Broyden':
		hartree_params, niter, alpha = pyqcm.broyden(F, X, iJ0 = alpha, maxiter=maxiter, convergence_test=G)
	elif iteration == 'fixed_point':
		hartree_params, niter = pyqcm.fixed_point_iteration(F, X, xtol=1e-6, convergence_test=G, maxiter=maxiter, alpha=alpha, eps_algo=eps_algo)
	else:
		raise ValueError('type of iteration unknown in call to Hartree_procedure(...)')

	I = pyqcm.model_instance(self)
	I.props['method'] = iteration
	I.props['iterations'] = niter
	if SEF: I.Potthoff_functional(hartree=couplings)
	I.write_summary(file)

	pyqcm.banner('Hartree procedure has converged', c='*')
	return I, niter, alpha
	


#---------------------------------------------------------------------------------------------------
# performs a loop over mu with adjustable step to make the density step more constant
# IN DEVELOPMENT
def flexible_loop(self, task, initial_mu, final_n, initial_step, delta_n=0.005, varia=[]):
	"""
	Performs a controlled loop for VCA or CDMFT with a predictor.  The definition of each
	model instance must be done and returned by 'task'; it is not done by this looping function.

	:param task: a function called at each step of the loop. Must return a model_instance.
	:param float initial_mu: initial value of the chemical potential mu
	:param float final_n: target final density at which the loop should stop
	:param float initial_step: initial step in mu for the first iteration
	:param float delta_n: approximate step in density to keep approximately constant across iterations
	:param [str] varia: names of the variational parameters
	:returns: None

	"""

	nvar = len(varia)  # number of variational parameters
	sol = np.empty((4, nvar))  # current solution + 3 previous solutions
	mu = np.empty(4)  # current mu + 3 previous values
	n = np.empty(4)  # current density + 3 previous values
	if initial_step*delta_n < 0 : delta_n *= -1

	mu[0] = initial_mu
	P = self.parameters()
	for i,v in enumerate(varia):
		sol[0] = P[v]
	loop_counter = 1
	
	while True:  # main loop here
		self.set_parameter('mu', mu[0])

		pyqcm.banner('loop index = {:d},  mu = {:1.4f}'.format(loop_counter + 1, mu[0]), '=', skip=1)
		try:
			I = task()

		except Exception:
			return
					
		n[0] = I.averages()['mu']
		print('density = ', n[0])
		mu = np.roll(mu, 1)  # cycles the values of the loop parameter
		n = np.roll(n, 1)  # cycles the values of the loop parameter
		sol = np.roll(sol, 1, 0)  # cycles the solutions


		n[0] = n[1] + delta_n
		if delta_n < 0 and n[0] < final_n: break
		if delta_n > 0 and n[0] > final_n: break

		# predicting the starting value from previous solutions (do nothing if loop_counter=0)
		fit_type = ''
		if loop_counter == 1:
			start = sol[1, :]  # the starting point is the previous value in the loop
			mu[0] = mu[1] + initial_step  # updates the loop parameter
		elif loop_counter == 2:
			fit_type = ' (linear fit)'
			x = mu[1:3]  # the two previous values of mu
			y = sol[1:3, :]  # the two previous values of the variational parameter
			z = n[1:3]  # the two previous values of the density
			pol_mu = np.polyfit(z, x, 1)
			pol = np.polyfit(x, y, 1)
			mu[0] = np.polyval(pol_mu, n[0])
			if abs(mu[0]-mu[1]) > 10*np.abs(initial_step):
				mu[0] = mu[1] + 10*initial_step
				loop_counter -= 1
			else:
				for i in range(nvar):
					start[i] = np.polyval(pol[:, i], mu[0])

		else:
			fit_type = ' (quadratic fit)'
			x = mu[1:4]  # the three previous values of the loop parameter
			y = sol[1:4, :]  # the three previous values of the loop parameter
			z = n[1:4]  # the three previous values of the density
			pol_mu = np.polyfit(z, x, 2)
			pol = np.polyfit(x, y, 2)
			mu[0] = np.polyval(pol_mu, n[0])
			for i in range(nvar):
				start[i] = np.polyval(pol[:, i], mu[0])

		if loop_counter > 1:
			print('flexible_loop step : mu = ', mu[0]-mu[1])
			print('predictor : ', start, fit_type)
			for i in range(len(varia)):
				self.set_parameter(varia[i], start[i])
				print(' ---> ', varia[i], ' = ', start[i])

		loop_counter += 1

	pyqcm.banner('flexible loop ended normally', '%')



import pyqcm
from pyqcm.cdmft import CDMFT
import numpy as np

import model_graphene_bath as M

# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
M.model.set_target_sectors(['R0:N6:S0'])

# Simulation parameters
M.model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    tb1_1=0.5
    tb2_1=0.4
    eb1_1=1.0
    eb2_1=-1.1
""")

varia = ['eb1_1', 'eb2_1', 'tb1_1', 'tb2_1']

# convergence = 'GS energy'; accur = 1e-3
# convergence='self-energy'; accur = 1e-4
# convergence='t'; accur = 1e-4
# convergence='parameters'; accur = 1e-4
# convergence=('parameters', 'hybridization', 'self-energy'); accur = (5e-4, 1e-4, 1e-4)
# convergence=('mu', 't'); accur=(1e-3, 1e-3)
# convergence=[]; accur = []
# convergence=('t',); accur=(1e-3,)
# convergence=['distance', 'self-energy', 'E0']; accur=[1e-3, 1e-4, 1e-7]
# convergence=['self-energy', 'E0']; accur=[1e-4, 1e-7]
convergence=['self-energy', 'distance']; accur=[1e-4, 1e-7]



X = CDMFT(M.model, varia=varia, wc=[0.5,5,5], grid_type='legendre', accur=1e-3, convergence='self-energy', miniter=1, maxiter=64, depth=1, iteration='fixed_point')
# X.I.plot_host_hybrid(np.linspace(0.01, 10, 100), (0,0)); exit()

# Defining a function that will run a cdmft procedure within controlled_loop()
def run_cdmft():
    try:
        alpha = X.alpha
    except:
        alpha = 0.0
    X = CDMFT(M.model, varia=varia, wc=10, selfnorm=True, accur=accur, convergence=convergence, converge_with_stdev=False, miniter=1, maxiter=64, depth=1, iteration='fixed_point', alpha=alpha)
    return X.I

# pyqcm.json_output = True
# Looping over values of U
M.model.controlled_loop(
    task=run_cdmft,
    varia=["tb1_1", "eb1_1"], 
    loop_param="U", 
    loop_range=(2, 4.1, 0.5)
)


import pyqcm
import pyqcm.vca as V
import matplotlib.pyplot as plt
import numpy as np

# ax = plt.gca()
ax = None
F = None

# print(dir(pyqcm.vca)); exit()
#-----------------------------------------------------------------

# plt.show()
##################################################################
# TEST UNITAIRE

def test_vca():
    x = None
    import model_2x2_C2_vca as M

    F = 'test_vca.pdf'

    pyqcm.banner('testing vca()', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 't_1'], start=(0.3, 1.3), accur=1e-3, steps=[5e-5, 5e-5], max=[10,10], method='SYMR1')


test_vca()

# transition
# transition_line
# vca_min


####################################################################################################
# Simple one-dimensional Hubbard model with spin-orbit coupling
# We construct a one-dimensional Hubbard model with the following non-interacting Hamiltonian, here expressed in $k$-space:
# $$
#     H_0 = \sum_k \mathcal{H}_k  ~~~~~ \mathcal{H}_k = -2t\cos k\;c^\dagger_k c_k - 2m \sin k\;c^\dagger_k \sigma_x c_k
# $$
# where $c_k$ is a two-component annihilation operator in spin space : $(c_{k,\uparrow}, c_{k,\downarrow})$ and $\sigma_x$ is a Pauli matrix.
# Plot the non-interacting $\left(U=0\right)$ dispersion of this model as a function of $k$ in the interval $[-\pi,\pi]$ for $t=1$ and $m=0.2$ and check that it agrees with your analytical computation.
####################################################################################################
import pyqcm
import numpy as np
import matplotlib.pyplot as plt

# declare a cluster model of 4 sites, named 'clus'
CM = pyqcm.cluster_model( 4)

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, ((0,0,0), (1,0,0), (2,0,0),(3,0,0))) 

# define a lattice model named '1D_4' made of the cluster(s) clus and superlattice vector (4,0,0)
model = pyqcm.lattice_model('1D_4', clus, ((4,0,0),))

# define a few operators in this model
model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)
# Defining the one-body Rashba coupling term
# tau=2 ---> imaginary contribution, ensures sin(k), sigma=1 ---> spin flip
model.hopping_operator("m", (1,0,0), -1, tau=2, sigma=1) 


# Note that spin is not conserved due to the Rashba term
model.set_target_sectors('R0:N4') 
model.set_parameters("""
    t=1
    m=0.4
""")

# Plotting the spectral function
I = pyqcm.model_instance(model)
I.spectral_function(wmax=2.5, path="line", nk=64, orb=1)


####################################################################################################
# Plotting the analytically-derived dispersion relation

# creating a grid of wavevectors in the Brillouin zone
k_grid = np.linspace(-np.pi, np.pi, 150)

t=1
m=0.4
# Defining the energy eigenvalues found analytically
def eig_1(k, t, m):
    return -2*(t*np.cos(k) + m*np.sin(k))

def eig_2(k, t, m):
    return -2*(t*np.cos(k) - m*np.sin(k))

# Plotting the analytic dispersion curve
plt.plot(k_grid, eig_1(k_grid, t, m), label="$\epsilon_1(k)$")
plt.plot(k_grid, eig_2(k_grid, t, m), label="$\epsilon_2(k)$")
plt.xlabel("k")
plt.ylabel("$\epsilon(k)$")
plt.xticks((-np.pi,0,np.pi), ("$-\pi$","$0$","$-\pi$"))
plt.legend()
plt.show()



"""Test SB-CDMFT for the 1x4 chain hybridized with two 4-orbital
subbaths each belonging to an irrep of C2 point group.
"""

import pyqcm
from pyqcm.cdmft import CDMFT

# Number of orbitals of the models
n_sites, n_baths = 4, 4
no = n_sites + n_baths

# The 2 coupling sequences correspond to the 2 irreps of C2
irreps = {
    "A": [1, 1, 1, 1],
    "B": [1, 1, -1, -1],
}
cluster_models = {}

for sb_name, irrep in irreps.items():
    cluster_model = pyqcm.cluster_model(
        n_sites, n_baths, name="1x4_4b_{:s}".format(sb_name)
    )
    for i in range(1, n_baths + 1):
        lab = i + n_sites
        cluster_model.new_operator(
            "eb{:d}".format(i), "one-body", [(lab, lab, 1), (lab + no, lab + no, 1)]
        )
        elem = [
            (1, lab, irrep[0]),
            (n_sites, lab, irrep[n_sites - 1]),
            (1 + no, lab + no, irrep[0]),
            (n_sites + no, lab + no, irrep[n_sites - 1]),
        ]
        cluster_model.new_operator("tb{:d}".format(i), "one-body", elem)
    cluster_models[sb_name] = cluster_model

cluster_model_A, cluster_model_B = cluster_models["A"], cluster_models["B"]

varia = []
for k in range(1, len(irreps) + 1):
    varia += ["eb{:d}_{:d}".format(i, k) for i in range(1, n_baths + 1)]
    varia += ["tb{:d}_{:d}".format(i, k) for i in range(1, n_baths + 1)]

clus = pyqcm.cluster(
    (cluster_model_A, cluster_model_B),
    [[i, 0, 0] for i in range(n_sites)],
)
model = pyqcm.lattice_model("1x4_4b_2sb", clus, [[n_sites, 0, 0]])

model.varia = varia
model.interaction_operator("U")
model.hopping_operator("t", (1, 0, 0), -1)

model.set_target_sectors(["R0:N8:S0", "R0:N8:S0"])

lattice_params = """
    U = 4
    t = 1
    mu = 2
"""

bath_params = """
    eb1_1 = 1.0
    eb2_1 = -1.0
    eb3_1 = 1.0
    eb4_1 = -1.0
    tb1_1 = 0.5
    tb2_1 = 0.5
    tb3_1 = 0.5
    tb4_1 = 0.5
    eb1_2 = 1.0
    eb2_2 = -1.0
    eb3_2 = 1.0
    eb4_2 = -1.0
    tb1_2 = 0.5
    tb2_2 = 0.5
    tb3_2 = 0.5
    tb4_2 = 0.5
"""

model.set_parameters(lattice_params + bath_params)

X = CDMFT(
    model,
    varia,
    accur_bath=1e-6,
    accur_dist=1e-12,
    method="BOBYQA",
    iteration="fixed_point",
)




import pyqcm
import numpy as np
from pyqcm.cdmft import CDMFT

from model_1D_2_4b import model

pyqcm.set_global_parameter('accur_OP', 1e-5)
# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
model.set_target_sectors(['R0:N6:S0'])

# Simulation parameters
model.set_parameters("""
    U=4
    mu=2
    t=1
    t_1=1e-9
    tb1_1=0.5
    tb2_1=1*tb1_1
    eb1_1=1.0
    eb2_1=-1.0*eb1_1
""")

# Defining a function that will run a cdmft procedure within controlled_loop()
def run_cdmft():
    X = CDMFT(model, varia=["tb1_1", "eb1_1"], accur=1e-4) 
    return X.I

def control(I):
    a = I.averages()
    if np.abs(a['mu']-1.0) > 0.01:
        return False
    else:
        return True

# Looping over values of U
model.controlled_loop(
    task=run_cdmft, 
    varia=["tb1_1", "eb1_1"], 
    loop_param="mu", 
    loop_range=(2, 1e-9, -0.02),
    control_func = control,
    predict = True,
    retry = 'skip'
)



from pyqcm import *
from pyqcm.cdmft import *
from pyqcm.spectral import *

import model_1D_2_4b

np.set_printoptions(precision=4, linewidth=512, suppress=True)

sec = 'R0:S0'

set_target_sectors([sec])
set_parameters("""
    U=4
    mu=1
    t=1
    tb1_1=0.5
    tb2_1=0.5
    eb1_1=1.5
    eb2_1=-1
    sb1_1 = 0.1
    sb2_1 = 0.1
    sbi1_1 = 1e-9
    sbi2_1 = 1e-9
""")
var = [
'eb1_1',
'eb2_1',
'tb1_1',
'tb2_1',
'sbi1_1',
'sbi2_1',
'sb1_1',
'sb2_1',
]

pyqcm.new_model_instance()
params = parameters() # params is a dict
nvar = len(var)
S = np.empty(nvar)
for i in range(nvar):
    S[i] = params[var[i]]

vset = [S]
print(vset)
cdmft_distance_debug(var, vset, beta=20, wc=1); exit()


# cluster_spectral_function(self=True, file = 's1.pdf')
# hybridization_function(wmax=4, file = 'h1.pdf')


from pyqcm import *
from pyqcm.Kolmogorov_Smirnov import *
import numpy as np
import pytest
from common import extract_GS_energy, extract_pole


format_to_test = ["S", "E"]

class Test_diagonalization:
    
    @pytest.mark.parametrize("fmt", format_to_test)
    def test_simulation(self,fmt):
        """
        Test if the simulation run
        """
        from scipy.optimize import minimize
        import model_2x2_2C_8b_C2v_L
        nb = 8
        set_global_parameter('verb_ED')
        set_global_parameter('Hamiltonian_format', fmt)
        set_target_sectors(['R0:N12:S0', 'R0:N8:S0'])
        set_parameters(
            """
            U = 14
            e = 2.3
            tpd = 2.1
            tpp = 1
            tppp = 0.2
            mu = 9.68
            eb1_1 = -0.006655534
            eb2_1 = -3.3502828
            eb3_1 = -4.1133163
            eb4_1 = -0.10235806
            eb5_1 = -4.113222
            eb6_1 = -0.1023468
            eb7_1 = 0.14108001
            eb8_1 = -3.7537133
            tb1_1 = -0.11240159
            tb2_1 = -1.1061747
            tb3_1 = 1.7476608
            tb4_1 = 0.057892245
            tb5_1 = -1.7476438
            tb6_1 = -0.057886658
            tb7_1 = 0.22749261
            tb8_1=-2.3061557
            """
        )
        I = new_model_instance(record=True)
        I.print(fmt+".out")
        return
    
    @pytest.mark.parametrize("fmt", format_to_test)
    def test_GS_energy(self,fmt):
        ref_gs_energy = extract_GS_energy("S_format.ref")
        test_gs_energy = extract_GS_energy(fmt+".out")
        print("Ref:", ref_gs_energy)
        print("Test:", test_gs_energy)
        assert np.isclose(ref_gs_energy["R0:N12:S0:1"], test_gs_energy["R0:N12:S0:1"])
    
    @pytest.mark.parametrize("fmt", format_to_test)
    def test_pole(self,fmt):
        """
        This test the results of the previous simulation against a reference file
        """
        M = extract_pole(fmt+".out")[0]
        print("Test:", M)
        M = M[M[:, 0].argsort(), :]
        M = (M[:,0],M[:,1]*M[:,1])
        R = extract_pole('S_format.ref')[0]
        print("Ref:", R)
        R = R[R[:, 0].argsort(), :]
        R = (R[:,0],R[:,1]*R[:,1])
        distance = Kolmogorov_Smirnov(M, R, tol=1e-6, plot=False)
        print('Kolmogorov-Smirnov distance = ', distance)
        assert distance < 0.01





import pyqcm
from model_graphene_bath import model

# print(model.clus[0].cluster_model.matrix_elements('tb1'))
print('matrix elements of eb1:')
print(model.systems[0].matrix_elements('eb1'))
print('matrix elements of tb1:')
print(model.systems[0].matrix_elements('tb1'))

model.set_parameters("""
U = 4
t = 1
eb1_1=1
tb1_1=0.5
eb2_1=-1*eb1_1
tb2_1=1*tb1_1
"""
)
model.set_target_sectors("N6:S0")
I = pyqcm.model_instance(model)

print('\nmatrix elements of the interactions:\n', I.interactions())

print('\nParameters:\n', I.parameters())

print('\ncluster hoppint matrix:\n', I.cluster_hopping_matrix())

print('\ncluster hoppint matrix (with bath):\n', I.cluster_hopping_matrix(0))

print('\nInstance parameters:\n', I.instance_parameters())



####################################################################################################
# Simple Hubbard model in d=1 with dimerization a cluster of 2 sites
####################################################################################################
import pyqcm

# declare a cluster model of 2 sites, named 'clus'
CM = pyqcm.cluster_model( 2)  

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, ((0,0,0), (1,0,0))) 

# define a lattice model named '1D_4' made of the cluster(s) clus and superlattice vector (2,0,0), same as the lattice vector
model = pyqcm.lattice_model('1D_4', clus, ((2,0,0),), ((2,0,0),))

# define a few operators in this model
model.interaction_operator('U', orbitals=(1,1))
model.interaction_operator('U', orbitals=(2,2))
model.hopping_operator('t', (1,0,0), -1, orbitals=(1,2))
model.hopping_operator('tp', (1,0,0), -1, orbitals=(2,1))


model.set_target_sectors(['R0:N2:S0']) # 2 particles, total spin 0
model.set_parameters("""
t=1.1
tp=0.9
U=1e-9
mu=0.5*U
""")


model.set_parameter("U", 1e-9) # To make sure correct U is used no matter execution order of cells
I = pyqcm.model_instance(model)
print(f"Ground state: {I.ground_state()}")
print(I.cluster_averages())
I.spectral_function(path="line") 


model.set_parameter('U', 4)
I = pyqcm.model_instance(model)
print(f"Ground state: {I.ground_state()}")
print(I.cluster_averages())
I.spectral_function(path="line") 



from pyqcm import *
from pyqcm.cdmft import *
from pyqcm.loop import *
from pyqcm.variable_parameter import *


import model_1D_2_4b

acc = 1e-3
np.set_printoptions(precision=4, linewidth=512, suppress=True)

def Uvsn(n):
    return 4 -2*(1-n)

Uvar = variable_parameter('U', Uvsn, obs='mu', lattice=False)

sec = 'R0:N6:S0/R0:N4:S0/R0:N2:S0'
set_target_sectors([sec])

set_parameters("""
    U=4
    mu=1
    t=1
    tb1_1=0.6
    tb2_1=0.5
    eb1_1=1.2
    eb2_1=-1
""")

varia=['tb1_1', 'tb2_1', 'eb1_1', 'eb2_1']

def F():
    cdmft(varia=varia, accur=1e-4, accur_hybrid=1e-8, accur_dist=1e-10, check_sectors=True)

variable_parameter_self_consistency(F, Uvar)



import pyqcm
from model_small import model

pyqcm.set_global_parameter('verb_integrals')
pyqcm.set_global_parameter('verb_Hilbert')
pyqcm.set_global_parameter('seed', 1234)
model.set_target_sectors('R0:S0')
model.set_parameters("""
t = 1
tz = 1
U = 1
mu = 0
D = 0.2
""")
I = pyqcm.model_instance(model)
I.averages(pr=True)




import pyqcm
from pyqcm.cdmft import CDMFT
from model_1D_2_4b import model

# import numpy as np
# A = np.kron([1,2], [10,20]).reshape((2,2))
# print(A); exit()

model.set_target_sectors('R0:N6:S0')
model.set_parameters("""
    U=4
    V=1
    Vm = 1
    mu=2
    t=1
    tb1_1=0.5
    tb2_1=0.55
    eb1_1=1.0
    eb2_1=-1.0
""")

Vm_H = pyqcm.hartree(model, 'Vm', 'V', model.Vm_eig, lattice=False) 
# Vm_H = pyqcm.hartree(model, 'Vm', 'V', model.Vm_eig, lattice=True) 

varia = ['eb1_1', 'eb2_1', 'tb1_1', 'tb2_1']
def run_cdmft(iteration):
    try:
        alpha = X.alpha
    except:
        alpha = 0.0
    U = model.parameters()['U']
    V = model.parameters()['V']
    model.set_parameter('mu', 0.5*U+2*V)
    try:
        X = CDMFT(model, varia, iteration=iteration, convergence=('GS energy', 'parameters'), accur=(1e-4, 1e-4), eps_algo=2, alpha = alpha, hartree=(Vm_H,)) 
        return X.I
    except:
        raise pyqcm.SolverError('Failure of the CDMFT method')

# Looping over values of U
model.controlled_loop(
    task=lambda : run_cdmft('fixed_point'),
    varia = varia,
    loop_param='U', 
    loop_range=(2, 4.1, 0.5)
)

model.controlled_loop(
    task=lambda : run_cdmft('broyden'), 
    varia = varia,
    loop_param='U', 
    loop_range=(2, 4.1, 0.5)
)


import numpy as np
import pyqcm
import matplotlib.pyplot as plt
import model_2x2_C2 as M

I = pyqcm.model_instance(M.model)

ax = None
F = None
# F = None
#-----------------------------------------------------------------
# import record_2x2_anom
# import record_spin

# DoS(w=4, eta=0.2, file = F, plt_ax=ax)
# Fermi_surface(file = F, plt_ax=ax)
# G_dispersion(max=20, contour=True, file = F, plt_ax=ax)
# Luttinger_surface(file = F, plt_ax=ax)
# cluster_spectral_function(file = F, plt_ax=ax)
# dispersion(file = F, contour=False, plt_ax=ax, labels=True)
# mdc(file = F, plt_ax=ax)
# momentum_profile('t', file = F, plt_ax=ax)
# segment_dispersion(file = F, plt_ax=ax)
# spectral_function(file = F, plt_ax=ax)
# spectral_function_Lehmann(lims=(-5,5), file = F, plt_ax=ax)
# spin_mdc(nk=20, orb=1, file = F, plt_ax=ax, opt='spins')
# spin_mdc(nk=200, orb=1, file = F, plt_ax=ax, opt='spinp')
# spin_mdc(nk=200, orb=1, file = F, plt_ax=ax, opt='sz')
# spin_mdc(nk=25, orb=1, file = F, plt_ax=ax, opt='spins')
# mdc_anomalous(orbs=(1,1), self=True, quadrant=True, file = F, plt_ax=ax)

#-----------------------------------------------------------------
# import record_2x2_8b

# hybridization_function(file = F, plt_ax=ax)

#-----------------------------------------------------------------
# import record_2x2_anom

# plt.show()


##################################################################
# TEST UNITAIRE
Lab = 9
def test_spectral():
    x = None
    # import record_2x2_anom
    # read_cluster_model_instance(record_2x2_anom.solution[0], 9)

    F = 'test_cluster_spectral_function.pdf'
    pyqcm.banner('testing cluster_spectral_function()', c='#', skip=1); I.cluster_spectral_function(file = F, plt_ax=ax)

    F = 'test_DoS.pdf'
    pyqcm.banner('testing DoS()', c='#', skip=1); I.plot_DoS(w=1, eta=0.2, file = F, plt_ax=ax)
    pyqcm.banner('testing DoS()', c='#', skip=1); I.plot_DoS(w=np.linspace(-1,1,40), eta=0.2, file = F, plt_ax=ax)

    F = 'test_G_dispersion.pdf'
    pyqcm.banner('testing G_dispersion()', c='#', skip=1); I.G_dispersion(max=20, file = F, plt_ax=ax)
    
    F = 'test_Luttinger_surface.pdf'
    pyqcm.banner('testing Luttinger_surface()', c='#', skip=1); I.Luttinger_surface(file = F, plt_ax=ax)
    
    F = 'test_mdc.pdf'
    pyqcm.banner('testing mdc()', c='#', skip=1); I.mdc(nk = 50, file = F, plt_ax=ax)
    
    F = 'test_momentum_profile.pdf'
    pyqcm.banner('testing momentum_profile()', c='#', skip=1); I.plot_momentum_profile('t', nk=10, file = F, plt_ax=ax)
    
    F = 'test_spectral_function.pdf'
    pyqcm.banner('testing spectral_function()', c='#', skip=1); I.spectral_function(nk=8, file = F, plt_ax=ax)
    F2 = 'test_spectral_function_M.pdf'
    pyqcm.banner('testing spectral_function()', c='#', skip=1); I.spectral_function(nk=8, file = F2, period='M', plt_ax=ax)
    F2 = 'test_spectral_function_N.pdf'
    pyqcm.banner('testing spectral_function()', c='#', skip=1); I.spectral_function(nk=8, file = F2, period='N', plt_ax=ax)
    F2 = 'test_spectral_function_self.pdf'
    pyqcm.banner('testing spectral_function()', c='#', skip=1); I.spectral_function(nk=8, file = F2, opt='self', offset=30, plt_ax=ax)

    F = 'test_spectral_function_Lehmann.pdf'
    pyqcm.banner('testing spectral_function_Lehmann()', c='#', skip=1); I.spectral_function_Lehmann(lims=(-5,5), nk=8, file = F, plt_ax=ax)
    
    if I.model.mixing&2:
        F = 'test_spin_mdc.pdf'
        pyqcm.banner('testing spin_mdc()', c='#', skip=1); I.spin_mdc(nk=40, opt='spins', file = F, plt_ax=ax)

    F = 'test_plot_hybridization_function.pdf'
    pyqcm.banner('testing plot_hybridization_function()', c='#', skip=1); I.plot_hybridization_function(file = F, plt_ax=ax)

    if I.model.mixing&1:
        F = 'test_mdc_anomalous.pdf'
        pyqcm.banner('testing mdc_anomalous()', c='#', skip=1); I.mdc_anomalous(nk=40, file = F, plt_ax=ax)

        F = 'test_plot_profile.pdf'
        pyqcm.banner('testing plot_profile()', c='#', skip=1); I.plot_profile(file = F)

    F = 'test_Fermi_surface.pdf'
    pyqcm.banner('testing Fermi_surface()', c='#', skip=1); I.Fermi_surface(file = F, plt_ax=ax)
    
    F = 'test_dispersion.pdf'
    pyqcm.banner('testing dispersion()', c='#', skip=1); I.plot_dispersion(file = F, plt_ax=ax)
    
    F = 'test_dispersionC.pdf'
    pyqcm.banner('testing dispersion() with contours', c='#', skip=1); I.plot_dispersion(contour=True, file = F, plt_ax=ax)

    F = 'test_segment_dispersion.pdf'
    pyqcm.banner('testing segment_dispersion()', c='#', skip=1); I.segment_dispersion(file = F, plt_ax=ax)
    

test_spectral()


import pyqcm
from model_1D import model1D
model = model1D(4)
import matplotlib.pyplot as plt
mixing = 3

if mixing == 4:
    model.set_target_sectors(['R0:N4:S0'])
    model.set_parameters("""
    U=1
    t=1
    H = -1
    mu = 1
    """)

elif mixing == 3:
    model.set_target_sectors(['R0'])
    model.set_parameters("""
    U=1
    t=1
    S = 0.4
    H = 0.5
    Hx = 1
    mu = 1
    """)

elif mixing == 2:
    model.set_target_sectors(['R0:N4'])
    model.set_parameters("""
    U=1
    t=1
    H = -1
    Hx = 1e-8 
    mu = 1
    """)

elif mixing == 1:
    model.set_target_sectors(['R0:S0'])
    model.set_parameters("""
    U=1
    t=1
    S = 0.4
    H = -1
    mu = 1
    """)

elif mixing == 0:
    model.set_target_sectors(['R0:N4:S0'])
    model.set_parameters("""
    U=1
    t=1
    mu = 1
    """)

I = pyqcm.model_instance(model)
I.spectral_function(wmax = 6, nk = 32, path='line', file='test_mixing_{:d}.pdf'.format(mixing))

I.cluster_spectral_function(wmax = 4, plt_ax=plt.gca())
pyqcm.set_global_parameter('continued_fraction')
model.set_parameter('U', 1)
I = pyqcm.model_instance(model)
I.cluster_spectral_function(wmax = 4, plt_ax=plt.gca(), color='r')
plt.savefig("test_mixing_cf.pdf")



####################################################################################################
# Simple Hubbard model in d=2 with superconductivity on a cluster of 4 sites
####################################################################################################
import pyqcm
import pyqcm.vca as vca
import numpy as np

# declare a cluster model of 4 sites, named 'clus'
CM = pyqcm.cluster_model( 4)

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, ((0,0,0), (1,0,0), (0,1,0),(1,1,0))) 

# define a lattice model named '2x2' made of the cluster(s) clus and superlattice vectors (2,0,0) & (0,2,0)
model = pyqcm.lattice_model('2x2', clus, ((2,0,0), (0,2,0)))

# define a few operators in this model
model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)
model.hopping_operator('t', (0,1,0), -1)
model.hopping_operator('tp', (1,1,0), -1)
model.hopping_operator('tp', (-1,1,0), -1)

# On-site s-wave
model.anomalous_operator('S', [0,0,0], 1)

# Extended s-wave
model.anomalous_operator('xS', [1,0,0], 1)
model.anomalous_operator('xS', [0,1,0], 1)

# d_{x^2-y^2}-wave
model.anomalous_operator('D', [1,0,0], 1)
model.anomalous_operator('D', [0,1,0], -1)

# d_{xy}-wave, similar as previous wave but defined with "diagonal" links
model.anomalous_operator('Dxy', [1,1,0], 1)
model.anomalous_operator('Dxy', [-1,1,0], -1)

# Defining the appropriate sector of Hilbert space
model.set_target_sectors('R0:S0')

# The first three parameters are used for the graph of the Potthoff functionnal
model.set_parameters("""
t=1
U=8
mu=1.2
S_1=1e-9
xS_1=1e-9
D_1=1e-9
Dxy_1=1e-9
""")

# Generating a grid for the Potthoff functionnals
grid_size = 25
Delta_grid = np.linspace(1e-9, 0.3, grid_size)
pyqcm.set_global_parameter('accur_SEF', 1e-5)
# The different types of superconductivity are explored by plotting the Potthoff functionnal for each variety
vca.plot_sef(model, 'S_1', Delta_grid)
model.set_parameter("S_1", 1e-9)

vca.plot_sef(model, 'xS_1', Delta_grid)
model.set_parameter('xS_1', 1e-9)

vca.plot_sef(model, 'D_1', Delta_grid)
model.set_parameter('D_1', 1e-9)

vca.plot_sef(model, 'Dxy_1', Delta_grid)
model.set_parameter('Dxy_1', 1e-9)




import pyqcm
from model_graphene_bath import model

import matplotlib.pyplot as plt


# model.draw_operator('t', show_neighbors=True, values=True, show_labels=True, plt_ax=plt.gca())
# plt.savefig('draw_t.pdf')
# plt.cla()

# model.draw_operator('U', show_neighbors=True, values=True, show_labels=True, plt_ax=plt.gca())
# plt.savefig('draw_U.pdf')
# plt.cla()

model.draw_operator('M', show_neighbors=True, values=True, show_labels=True, plt_ax=plt.gca())
plt.savefig('draw_M.pdf')
plt.cla()

model.draw_operator('Mx', show_neighbors=True, values=True, show_labels=True, plt_ax=plt.gca())
plt.savefig('draw_Mx.pdf')
plt.cla()

model.draw_operator('cdw', show_neighbors=True, values=True, show_labels=True, plt_ax=plt.gca())
plt.savefig('draw_cdw.pdf')
plt.cla()

model.draw_cluster_operator(0, 'tb1', values=True, show_labels=True, plt_ax=plt.gca())
plt.savefig('draw_tb1.pdf')





####################################################################################################
# Simple Hubbard model on the graphene lattice
####################################################################################################
import pyqcm
import numpy as np

# declare a cluster model of 6 sites, named 'clus'
CM = pyqcm.cluster_model( 6)

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, ((1,0,0), (1,1,0), (0,1,0), (-1,0,0), (-1,-1,0), (0,-1,0))) 

# define a lattice model named '2x2' made of the cluster(s) clus and superlattice vectors (3,0,0) & (0,3,0)
# the lattice vectors are (1,-1,0) and (2,1,0), different from the default ones (1,0,0) & (0,1,0)
model = pyqcm.lattice_model('graphene6', clus, ((3,0,0), (0,3,0)), ((1,-1,0), (2,1,0)))

# we define a physical basis (expression of the working basis in terms of physical coordinates), for plotting purposes
model.set_basis(((1,0,0), (-0.5,np.sqrt(3)/2,0)))


# define a few operators in this model. Since this is a 2-orbital model, one needs to specify the relevant orbitals.
model.interaction_operator('U', orbitals=(1,1))
model.interaction_operator('U', orbitals=(2,2))
model.hopping_operator('t', (1,0,0), -1, orbitals=(1,2))
model.hopping_operator('t', (0,1,0), -1, orbitals=(1,2))
model.hopping_operator('t', (-1,-1,0), -1, orbitals=(1,2))


# (optional) check the definition of an operator by drawing it on the screen
model.draw_operator('t')

model.set_target_sectors('R0:N6:S0')
model.set_parameters("""
t=1
U=2
mu=1
""")

# define an instance of the model with those values
I = pyqcm.model_instance(model)

# solve for the ground state and prints the ground state averages
print('ground_state sector:', I.ground_state())
I.cluster_averages(pr=True)

# draws the spectral function of the cluster model
I.cluster_spectral_function(wmax=6)

# draws the spectral function of the lattice model, for wavevectors along a line from -pi to pi
I.spectral_function(wmax=6, path='graphene')

# draws a Fermi surface plot (spectral function as a function of wavevector, at zero frequency)
I.mdc(eta=0.05)

# draws the non-interacting dispersion relation
I.plot_dispersion()

# draws the density of states of the lattice model, from w = -6 to 6
I.plot_DoS(w = 6)



from pyqcm import *

new_cluster_model('L4', 4, 0, [[4,3,2,1]])
add_cluster('L4', [0, 0, 0], [[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0]])
lattice_model('1D_L4', [[4, 0, 0]])

interaction_operator('U')
hopping_operator('t', [1, 0, 0], -1)  # NN hopping
hopping_operator('tp', [2, 0, 0], -1)  # NNN hopping
hopping_operator('hx', [0, 0, 0], 1, tau=0, sigma=1)  # field in the x direction
hopping_operator('h', [0, 0, 0], 1, tau=0, sigma=3)  # field in the x direction



import pyqcm
import model_1D_2_4b as model

sec = 'R0:N6:S0'
model.model.set_target_sectors([sec])
model.model.set_parameters("""
t=1
U = 4
mu = 0.5*U
eb1_1 = 0.5
eb2_1 = -0.5
tb1_1 = 0.5
tb2_1 = 0.5
""")

I = pyqcm.model_instance(model.model)
I.cluster_averages(pr=True)

ave = I.Green_function_average()
print('\naverages of c^\dagger_i c_j :\n\n', ave)

print('\naverage of t from GF= ', -2*(ave[0,1]))
print('average of mu from GF = ', I.Green_function_density())






import pyqcm
from pyqcm.cdmft import CDMFT

from model_graphene_bath import model

# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
model.set_target_sectors(['R0:N6:S0'])

# Simulation parameters
model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    t_1=1e-9
    tb1_1=0.5
    tb2_1=1*tb1_1
    eb1_1=1.0
    eb2_1=-1.0*eb1_1
""")

I = pyqcm.model_instance(model)                
A = I.cluster_hopping_matrix()
print(A)



import pyqcm

CM = pyqcm.cluster_model(2)
clus = pyqcm.cluster(CM,((0,0,0),(1,0,0)))
model = pyqcm.lattice_model('1D_L2', clus, ((2,0,0),))

model.interaction_operator('U')
model.interaction_operator('J', link=(1,0,0), type='Hund')
model.hopping_operator('t', (1,0,0), -1)  # NN hopping
model.hopping_operator('tsi', (1,0,0), -1, sigma=3, tau=2)  # NN hopping
model.hopping_operator('ti', (1,0,0), -1, tau=2)  # NN hopping with imaginary amplitude
model.hopping_operator('tp', (2,0,0), -1)  # NNN hopping
model.hopping_operator('hx', (0,0,0), 1, tau=0, sigma=1)  # field in the x direction
model.hopping_operator('h', (0,0,0), 1, tau=0, sigma=3)  # field in the x direction
model.anomalous_operator('D', (1,0,0), 1)  # NN singlet
model.anomalous_operator('Di', (1,0,0), 1j)  # NN singlet with imaginary amplitude
model.anomalous_operator('S', (0,0,0), 1)  # on-site singlet
model.anomalous_operator('Si', (0,0,0), 1j)  # on-site singlet with imaginary amplitude
model.anomalous_operator('Pz', (1,0,0), 1, type='dz')  # NN triplet
model.anomalous_operator('Py', (1,0,0), 1, type='dy')  # NN triplet
model.anomalous_operator('Px', (1,0,0), 1, type='dx')  # NN triplet
model.density_wave('M', 'Z', (1,0,0))
model.density_wave('H', 'Z', (0,0,0))
model.density_wave('Hx', 'X', (0,0,0))



import numpy as np


def extract_GS_energy(f_in):
    src = open(f_in, 'r')
    gs_energy = {} #sector, energy
    for line in src.readlines():
        if 'GS_energy:' in line:
            data = line.split()
            gs_energy[data[-1]] = float(data[1])
    src.close()
    return gs_energy

def extract_pole(f_in):
    src = open(f_in, 'r')
    lines = src.readlines()
    src.close()
    data = []
    for i,line in enumerate(lines):
        if line[0:2] == "w\t":
            n_pole = int(line.split()[-1])
            values = [np.fromstring(x,sep="\t",dtype='f8').tolist() for x in lines[i+1:i+n_pole+1]]
            data.append(np.array(values, dtype='f8'))
    return data



"""Interface utilities for external exact-diagonalization (ED) solvers.

Provides :func:`freq_grid` for building Matsubara frequency grids and
:func:`test_ed_solver` for writing impurity problem data to a file and
reading back the resulting cluster model instance from an external solver.
"""
import pyqcm
import os
import subprocess
import numpy as np
import pyqcm.cdmft as cdmft

def freq_grid(beta=50,wc=2,real=False):
    """
    Builds a Matsubara frequency grid up to a cutoff frequency.

    :param float beta: inverse fictitious temperature used to define the Matsubara frequency spacing (2*pi/beta)
    :param float wc: cutoff frequency; the grid runs from pi/beta to wc
    :param bool real: if False (default), returns purely imaginary frequencies (multiplied by 1j); if True, returns real-valued frequencies
    :returns: array of frequencies (complex if real=False, real otherwise)
    :rtype: numpy.ndarray
    """
    wr = np.arange((np.pi / beta), wc + 1e-6, 2 * np.pi / beta)
    w = np.ones(len(wr), dtype=np.complex128)
    if real == False: 
        w = np.ones(len(wr), dtype=np.complex128)
        w = w * 1j
    else:
        w = np.ones(len(wr))
    w *= wr
    return w

def test_ed_solver():
    """
    Tests an external ED (exact diagonalization) solver by writing the impurity problem data
    (hopping matrix, interaction matrix, hybridization function, bath parameters) to a file
    named 'test.tsv', invoking an external Python script 'test_ed_solver.py', and then
    reading back the resulting cluster model instance from 'instance.out'.
    """

    info = pyqcm.cluster_info()[0]
    n = info[1]
    nn = 2*n
    nb = info[2]
    no = n + nb
    T = np.real(pyqcm.cluster_hopping_matrix(clus=0, spin_down=False))
    I = pyqcm.interactions()
    V = np.zeros((n,n))
    for x in I:
        if x[0] < n and x[1] < n: V[x[0], x[1]] = x[2]
    for x in I:
        if x[0] < n and x[1] == x[0]+no: V[x[0], x[0]] = x[2]

    w = freq_grid(beta=100,wc=10,real=True)

    P = pyqcm.parameters()

    with open('test.tsv', 'w') as f:
        f.write('Tij\n')
        np.savetxt(f, T, fmt='%1.6g', delimiter='\t')
        f.write('\n')
        f.write('Vij\n')
        np.savetxt(f, V, fmt='%1.6g', delimiter='\t')
        f.write('\n')
        f.write('Delta\n')
        for x in w:
            f.write(f'{x:1.6g}\t')
            D = pyqcm.hybridization_function(0,x*1j)
            D = D.reshape((1,nn))
            np.savetxt(f, D, fmt='%-1.6g', delimiter='\t')
        f.write('\n')
        f.write('bath_parameters\n')
        for x in cdmft.var:
            f.write(x+'\t{:1.6g}\n'.format(P[x]))
        f.write('\n')

    # now the file is written, one can call the external solver
    out = os.system("python test_ed_solver.py")
    if out != 0:
        raise ValueError('The external solver returned an error')

    # then we need to read the Green function representation to keep going
    # file_Green = 'Green.out'
    # if not os.path.isfile(file_Green):
    #     raise FileNotFoundError('The file ' + file_Green + ' does not exist!')
    # with open(file_Green) as f:
    #     solution = f.read()
    
    
    pyqcm.read_cluster_model_instance('instance.out', 0)





####################################################################################################
# CDMFT and the Mott transition on the graphene lattice
# We construct a model for the Hubbard model on the graphene lattice, based on a two-site cluster with a bath of 4 sites.
# We check that the correct dispersion relation is found at $U=0$ by plotting the spectral function.
# At half filling, we find the CDMFT solutions (normal state) as a function of $U$, from $U=10$ to $U=0$ in steps of $\Delta U=0.2$. We stay at half-filling, using particle-hole symmetry.
####################################################################################################
import pyqcm
import pyqcm.cdmft as cdmft
import numpy as np
import os

# declare a cluster model of 6 sites, named 'clus'
ns = 2 # number of physical sites
nb = 4 # number of bath sites
no = ns+nb # total number of sites
CM = pyqcm.cluster_model( ns, nb)

# Defining the bath hopping operators
CM.new_operator('tb1', 'one-body', [
    (1, 3, -1.0),
    (2, 4, -1.0),
    (1+no, 3+no, -1.0),
    (2+no, 4+no, -1.0)
]) # note that the last two entries pertain to the SPIN DOWN part of the operator

CM.new_operator('tb2', 'one-body', [
    (1, 5, -1.0),
    (2, 6, -1.0),
    (1+no, 5+no, -1.0),
    (2+no, 6+no, -1.0)
])

# Defining the 'orbital energy' of the baths
CM.new_operator('eb1', 'one-body', [
    (3, 3, 1.0),
    (4, 4, 1.0),
    (3+no, 3+no, 1.0),
    (4+no, 4+no, 1.0)
])

CM.new_operator('eb2', 'one-body', [
    (5, 5, 1.0),
    (6, 6, 1.0),
    (5+no, 5+no, 1.0),
    (6+no, 6+no, 1.0)
])

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, ((0,0,0), (1,0,0))) 

# define a lattice model named '2x2' made of the cluster(s) clus and superlattice vectors (3,0,0) & (0,3,0)
# the lattice vectors are (1,-1,0) and (2,1,0), different from the default ones (1,0,0) & (0,1,0)
model = pyqcm.lattice_model('graphene2_4b', clus, ((1,-1,0), (2,1,0)), ((1,-1,0), (2,1,0)))

# we define a physical basis (expression of the working basis in terms of physical coordinates), for plotting purposes
model.set_basis(((1,0,0), (-0.5,np.sqrt(3)/2,0)))


# define a few operators in this model. Since this is a 2-orbital model, one needs to specify the relevant orbitals.
model.interaction_operator('U', orbitals=(1,1))
model.interaction_operator('U', orbitals=(2,2))
model.hopping_operator('t', (1,0,0), -1, orbitals=(1,2))
model.hopping_operator('t', (0,1,0), -1, orbitals=(1,2))
model.hopping_operator('t', (-1,-1,0), -1, orbitals=(1,2))


model.set_target_sectors(['R0:N6:S0'])
model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    tb1_1=0.5
    tb2_1=1*tb1_1
    eb1_1=1.0
    eb2_1=-1.0*eb1_1
""")

# Setting a trivially small interaction value and plotting the spectral function
model.set_parameter("U", 1e-9)
I = pyqcm.model_instance(model)
I.spectral_function(path="graphene", orb=1) # Using the built-in graphene path is recommended here

# Setting up the range of values of U over which to loop the CDMFT simulation 
U_start = 0.2 # starting at U=0.2 because the nearly non-interacting case is too long to calculate and useless
U_stop = 10.1
U_step = 0.2

# Removes the default output file for display purposes
if os.path.isfile("./cdmft.tsv"):
    os.remove("./cdmft.tsv")

# Defining a function that will run a cdmft procedure within controlled_loop()
varia=('tb1_1', 'eb1_1')
def run_cdmft():
    U = model.parameters()['U']
    cdmft_simulation = cdmft.CDMFT(model, varia=varia, wc=10, grid_type='self', beta=50)
    return cdmft_simulation.I

# Looping over values of U
model.controlled_loop(
    task=run_cdmft, 
    varia=varia, 
    loop_param='U', 
    loop_range=(U_start, U_stop, U_step)
)


# plotting the results
import matplotlib.pyplot as plt

cdmft_data = np.genfromtxt("cdmft.tsv", names=True)

# plotting both orbital energies and site<-->bath hopping operator amplitude
ax1 = plt.subplot(211)
ax1.plot(cdmft_data['U'], cdmft_data['eb1_1'], "o", markersize=2, label="eb1_1")
ax1.set_ylabel("$\epsilon$")
ax1.legend()

ax2 = plt.subplot(212, sharex=ax1)
ax2.plot(cdmft_data['U'], cdmft_data['tb1_1'], "o", markersize=2, label="tb1_1")
ax2.set_xlabel("U")
ax2.set_ylabel("t")
ax2.legend()

plt.show()


index_of_U = np.where(cdmft_data['U'] == 4)[0][0] # Finding the index at which U=4
print(index_of_U)
model.set_params_from_file('cdmft.tsv', n=index_of_U)
I = pyqcm.model_instance(model)
I.spectral_function(path="graphene", orb=1) 

index_of_U = np.where(cdmft_data['U'] == 5)[0][0] # Finding the index at which U=4
model.set_params_from_file('cdmft.tsv', n=index_of_U)
I = pyqcm.model_instance(model)
I.spectral_function(path="graphene", orb=1) 

index_of_U = np.where(cdmft_data['U'] == 6)[0][0] # Finding the index at which U=4
model.set_params_from_file('cdmft.tsv', n=index_of_U)
I = pyqcm.model_instance(model)
I.spectral_function(path="graphene", orb=1) 



from pyqcm import *
set_global_parameter('verb_warning')

new_cluster_model('L4', 4, 0)
add_cluster('L4', [0, 0, 0], [[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0]])
lattice_model('1D_L4', [[4, 0, 0]])

interaction_operator('U')
interaction_operator('V', link=[1,0,0], amplitude=1)
interaction_operator('J', link=[1,0,0], type='Hund')
hopping_operator('t', [1, 0, 0], -1)  # NN hopping
hopping_operator('ti', [1, 0, 0], -1, tau=2)  # NN hopping with imaginary amplitude
hopping_operator('tp', [2, 0, 0], -1)  # NNN hopping
hopping_operator('hx', [0, 0, 0], 1, tau=0, sigma=1)  # field in the x direction
hopping_operator('h', [0, 0, 0], 1, tau=0, sigma=3)  # field in the x direction
anomalous_operator('D', [1, 0, 0], 1)  # NN singlet
anomalous_operator('Di', [1, 0, 0], 1j)  # NN singlet with imaginary amplitude
anomalous_operator('S', [0, 0, 0], 1)  # on-site singlet
anomalous_operator('Si', [0, 0, 0], 1j)  # on-site singlet with imaginary amplitude
anomalous_operator('Pz', [1, 0, 0], 1, type='dz')  # NN triplet
anomalous_operator('Py', [1, 0, 0], 1, type='dy')  # NN triplet
anomalous_operator('Px', [1, 0, 0], 1, type='dx')  # NN triplet
density_wave('M', 'Z', [1, 0, 0])
density_wave('H', 'Z', [0, 0, 0])
density_wave('Hx', 'X', [0, 0, 0])
density_wave('cdw', 'N', [1, 0, 0])

e = np.sqrt(0.5)
explicit_operator('V0m', [
([0,0,0], [0,0,0], e),
([3,0,0], [0,0,0], e),
], tau=0, type='one-body')

explicit_operator('V1m', [
([0,0,0], [0,0,0], e),
([3,0,0], [0,0,0],-e),
], tau=0, type='one-body')

# from pyqcm.draw_operator import draw_operator
# draw_operator('t')


import pyqcm

CM = pyqcm.cluster_model(4, generators=[[4, 3, 2, 1]])
clus = pyqcm.cluster(CM, ((0, 0, 0), ( 1, 0, 0), ( 0, 1, 0), ( 1, 1, 0)))
model = pyqcm.lattice_model('2x2_C2', clus, [( 2, 0, 0), ( 0, 2, 0)])
model.interaction_operator('U')
model.interaction_operator('V', link=( 1, 0, 0), amplitude=1)
model.interaction_operator('V', link=( 0, 1, 0), amplitude=1)
model.hopping_operator('Vm', [0, 0, 0], 1, tau=0)  # NN hopping
V_eig = 0.5
model.interaction_operator('J', link=( 1, 0, 0), type='Hund')

model.hopping_operator('t', ( 0, 1, 0), -1)
model.hopping_operator('t', ( 1, 0, 0), -1)
model.hopping_operator('t2', ( 1, 1, 0), -1)
model.hopping_operator('t2', (-1, 1, 0), -1)
model.hopping_operator('hx', [0, 0, 0], -1, tau=0, sigma=1)
model.density_wave('M', 'Z', ( 1, 1, 0))
model.hopping_operator('t3', ( 2, 0, 0), -1)
model.hopping_operator('t3', ( 0, 2, 0), -1)
model.anomalous_operator('D', ( 1, 0, 0), 1)
model.anomalous_operator('D', ( 0, 1, 0), -1)


################################################
model.set_target_sectors(['R0:N4:S0'])
model.set_parameters("""
t = 1
t2 = 1e-9
U = 8
mu = 4
mu_1 = 4
M = 0
M_1 = 0.15
t_1 = 1
# D = 0.001
# hx = 0.001
""")
                     



import pyqcm
from model_1D import model1D

L=6
model = model1D(L)
model.set_target_sectors('R0:N{:d}:S0'.format(L))

model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
""")

I = pyqcm.model_instance(model)
print('GF written:\n', I.cluster_Green_function(0.1j, 0))
I.write_hdf5('test.h5')
I = pyqcm.model_instance(model)
I.read_hdf5('test.h5', set_parameters=True)
print('\nGF read:\n', I.cluster_Green_function(0.1j, 0))



from pyqcm import *

new_cluster_model('clus', 4, 8, [[2, 1, 4, 3, 0, 0, 0, 0, 2, 2, 2, 2], [3, 4, 1, 2, 0, 0, 2, 2, 0, 0, 2, 2]], bath_irrep=True)
new_cluster_model('clus_uncorrelated',8,0)


# bath energies
for i in range(1,9):
    name = 'eb'+str(i)
    lab = i+4
    new_cluster_operator('clus', name, 'one-body', [
        (lab, lab, 1.0),
        (lab + 12, lab + 12, 1.0)
    ])
    

def new_tb(x, seq):
    elem = []
    for i in range(4):
        elem.append((i+1, x+4, seq[i]))
    for i in range(4):
        elem.append((i+13, x+16, seq[i]))
    name = 'tb'+str(x)
    new_cluster_operator('clus', name, 'one-body', elem)

new_tb(1, [1, 1, 1, 1])
new_tb(2, [1, 1, 1, 1])
new_tb(3, [1, 1, -1, -1])
new_tb(4, [1, 1, -1, -1])
new_tb(5, [1, -1, 1, -1])
new_tb(6, [1, -1, 1, -1])
new_tb(7, [1, -1, -1, 1])
new_tb(8, [1, -1, -1, 1])

def new_pairing(x, seq):
    elem = []
    for i in range(4):
        elem.append((i+1, x+4+12, seq[i]))
    for i in range(4):
        elem.append((x+4, i+1+12, seq[i]))
    name = 'sb'+str(x)
    new_cluster_operator('clus', name, 'anomalous', elem)

new_pairing(1, [1, 1, 1, 1])
new_pairing(2, [1, 1, 1, 1])
new_pairing(3, [1, 1, -1, -1])
new_pairing(4, [1, 1, -1, -1])
new_pairing(5, [1, -1, 1, -1])
new_pairing(6, [1, -1, 1, -1])
new_pairing(7, [1, -1, -1, 1])
new_pairing(8, [1, -1, -1, 1])







import pyqcm

def model2D(Lx, Ly, sym=True):

    gen1 = []
    for y in range(Ly):
        gen1 += [x+y*Lx for x in range(Lx,0,-1)]
    gen2 = []
    for y in range(Ly-1, -1, -1):
        gen2 += [x+y*Lx for x in range(1, Lx+1)]

    # print(gen1)
    # print(gen2)
    # exit()

    if sym:
        CM = pyqcm.cluster_model(Lx*Ly, 0, generators=(gen1, gen2))
    else:
        CM = pyqcm.cluster_model(Lx*Ly, 0)
       
    pos = [(x,y,0) for y in range(Ly) for x in range(Lx) ]

    clus = pyqcm.cluster(CM, pos)
    model = pyqcm.lattice_model('rect_{:d}x{:d}'.format(Lx,Ly), clus, ((Lx,0,0),(0,Ly,0)))

    model.interaction_operator('U')
    model.hopping_operator('t', (1,0,0), -1)
    model.hopping_operator('t', (0,1,0), -1)

    return model